import unittest
from os.path import join
from context import grama as gr
from context import models

## Core function tests
##################################################
class TestPipe(unittest.TestCase):
    def setUp(self):
        self.md = models.make_test()

    def test_pipe(self):
        ## Chain
        res = self.md >> gr.ev_hybrid(df_det="nom") >> gr.tf_sobol()

class TestFindFiles(unittest.TestCase):
    def test_find_files(self):
        # Non-recursive
        res = gr.find_files(".", ".py", recursive=False)
        self.assertTrue(
            join(".", "context.py") in res
        )
        self.assertTrue(
            not join("longrun", "sp_convergence.ipynb") in res
        )
        # Handles missing ext dot
        res = gr.find_files(".", "py", recursive=False)
        self.assertTrue(
            join(".", "context.py") in res
        )
        # No return
        res = gr.find_files(".", ".exe", recursive=False)
        self.assertTrue(len(res) == 0)
        # Recursive
        res = gr.find_files(".", ".ipynb", recursive=True)
        self.assertTrue(
            join(".", "longrun", "sp_convergence.ipynb") in res
        )
        # Accepts paths to join
        res = gr.find_files([".", "longrun"], ".ipynb", recursive=False)
        self.assertTrue(
            join(".", "longrun", "sp_convergence.ipynb") in res
        )



import numpy as np
import pandas as pd
import unittest
import io
import sys

from context import grama as gr

## Test cohort shapley
##################################################
class TestCohortShapley(unittest.TestCase):
    def setUp(self):
        pass

    def test_cohort_shapley(self):
        df_data = gr.df_make(x0=[0, 0, 1, 1], x1=[0, 1, 0, 1], f=[0, 1, 1, 2],)

        df_true = gr.df_make(
            f_x0=[-0.5, -0.5, +0.5, +0.5], f_x1=[-0.5, +0.5, -0.5, +0.5],
        )

        df_cohort = gr.tran_shapley_cohort(df_data, var=["x0", "x1"], out=["f"])

        self.assertTrue(gr.df_equal(df_true, df_cohort))



import numpy as np
import scipy.linalg
from .misc import check_sample_inputs
from .local_linear import local_linear_grads  

def initialize_subspace(X = None, fX = None, grads = None, n_grads = 100):
	r""" Construct an initial estimate of the desired subspace 
	"""


	X, fX, grads = check_sample_inputs(X, fX, grads)
	ngrads = max(n_grads, X.shape[1])
	# If we don't have enough grads and we have enough samples to estimate gradients
	if len(grads) < n_grads and X.shape[0] >= X.shape[1]+1:
		# Pick a random subset to estimate the gradient at
		I = np.random.permutation(X.shape[0])[:n_grads - len(grads)]
		grad_est = local_linear_grads(X, fX, Xt = X[I])
		all_grads = np.vstack([grads, grad_est])
	else:
		all_grads = grads

	# Compute SVD
	U, s, VH = scipy.linalg.svd(all_grads.T, full_matrices = False, compute_uv = True)
	return U	
	



""" Descriptions of various bases"""

__all__ = ['PolynomialTensorBasis',
	'MonomialTensorBasis',
	'LegendreTensorBasis',
	'ChebyshevTensorBasis',
	'LaguerreTensorBasis',
	'HermiteTensorBasis',
	'ArnoldiPolynomialBasis',
 ]

import numpy as np
from numpy.polynomial.polynomial import polyvander, polyder, polyroots
from numpy.polynomial.legendre import legvander, legder, legroots
from numpy.polynomial.chebyshev import chebvander, chebder, chebroots
from numpy.polynomial.hermite import hermvander, hermder, hermroots
from numpy.polynomial.laguerre import lagvander, lagder, lagroots

from itertools import product

class Basis(object):
	pass



################################################################################
# Indexing utility functions for total degree
################################################################################

#TODO: Should these be moved into PolynomialTensorBasis class?

def _full_index_set(n, d):
	""" A helper function for index_set.

	Parameters
	----------
	n : int
		degree of polynomial
	d : int
		number of variables, dimension
	"""
	if d == 1:
		I = np.array([[n]])
	else:
		II = _full_index_set(n, d-1)
		m = II.shape[0]
		I = np.hstack((np.zeros((m, 1)), II))
		for i in range(1, n+1):
			II = _full_index_set(n-i, d-1)
			m = II.shape[0]
			T = np.hstack((i*np.ones((m, 1)), II))
			I = np.vstack((I, T))
	return I

def index_set(n, d):
	"""Enumerate multi-indices for a total degree of order `n` in `d` variables.

	Parameters
	----------
	n : int
		degree of polynomial
	d : int
		number of variables, dimension
	Returns
	-------
	I : ndarray
		multi-indices ordered as columns
	"""
	I = np.zeros((1, d), dtype = int)
	for i in range(1, n+1):
		II = _full_index_set(i, d)
		I = np.vstack((I, II))
	return I[:,::-1].astype(int)


class PolynomialTensorBasis(Basis):
	r""" Generic tensor product basis of fixed total degree

	This class constructs a tensor product basis of dimension :math:`n`
	of fixed given degree :math:`p` given a basis for polynomials
	in one variable. Namely, this basis is composed of elements:

	.. math::

		\psi_j(\mathbf x) := \prod_{i=1}^n \phi_{[\boldsymbol \alpha_j]_i}(x_i)
			\quad \sum_{i=1}^n [\boldsymbol \alpha_j]_i \le p;
			\quad \phi_i \in \mathcal{P}_{i}(\mathbb{R})


	Parameters
	----------
	dim: int
		The input dimension of the space
	degree: int
		The total degree of polynomials
	polyvander: function
		Function providing the scalar Vandermonde matrix (i.e., numpy.polynomial.polynomial.polyvander)
	polyder: function
		Function providing the derivatives of scalar polynomials (i.e., numpy.polynomial.polynomial.polyder)

	"""

	def __init__(self, degree, X = None, dim = None):
		self.degree = int(degree)
		if X is not None:
			self.X = np.atleast_2d(X)
			self.dim = self.X.shape[1]
			self.set_scale(self.X)
		elif dim is not None:
			self.dim = int(dim)
			self.X = None

		self.indices = index_set(self.degree, self.dim).astype(int)
		self._build_Dmat()

	def __len__(self):
		return len(self.indices)

	def _build_Dmat(self):
		""" Constructs the (scalar) derivative matrix
		"""
		self.Dmat = np.zeros( (self.degree+1, self.degree))
		I = np.eye(self.degree + 1)
		for j in range(self.degree + 1):
			self.Dmat[j,:] = self.polyder(I[:,j])

	def set_scale(self, X):
		r""" Construct an affine transformation of the domain to improve the conditioning
		"""
		self._set_scale(np.array(X))

	def _set_scale(self, X):
		r""" default scaling to [-1,1]
		"""
		self._lb = np.min(X, axis = 0)
		self._ub = np.max(X, axis = 0)

	def _scale(self, X):
		r""" Apply the scaling to the input coordinates
		"""
		try:
			return 2*(X-self._lb[None,:])/(self._ub[None,:] - self._lb[None,:]) - 1
		except AttributeError:
			return X

	def _dscale(self):
		r""" returns the scaling associated with the scaling transform
		"""
		try:
			return (2./(self._ub - self._lb))
		except AttributeError:
			raise NotImplementedError

	def V(self, X = None):
		r""" Builds the Vandermonde matrix associated with this basis

		Given points :math:`\mathbf x_i \in \mathbb{R}^n`,
		this creates the Vandermonde matrix

		.. math::

			[\mathbf{V}]_{i,j} = \phi_j(\mathbf x_i)

		where :math:`\phi_j` is a multivariate polynomial as defined in the class definition.

		Parameters
		----------
		X: array-like (M, n)
			Points at which to evaluate the basis at where :code:`X[i]` is one such point in
			:math:`\mathbf{R}^n`.

		Returns
		-------
		V: np.array
			Vandermonde matrix
		"""
		if X is None and self.X is not None:
			X = self.X
		elif X is None:
			raise NotImplementedError

		X = X.reshape(-1, self.dim)
		X = self._scale(np.array(X))
		M = X.shape[0]
		assert X.shape[1] == self.dim, "Expected %d dimensions, got %d" % (self.dim, X.shape[1])
		V_coordinate = [self.vander(X[:,k], self.degree) for k in range(self.dim)]

		V = np.ones((M, len(self.indices)), dtype = X.dtype)

		for j, alpha in enumerate(self.indices):
			for k in range(self.dim):
				V[:,j] *= V_coordinate[k][:,alpha[k]]
		return V

	def VC(self, X, c):
		r""" Evaluate the product of the Vandermonde matrix and a vector

		This evaluates the product :math:`\mathbf{V}\mathbf{c}`
		where :math:`\mathbf{V}` is the Vandermonde matrix defined in :code:`V`.
		This is done without explicitly constructing the Vandermonde matrix to save
		memory.

		Parameters
		----------
		X: array-like (M,n)
			Points at which to evaluate the basis at where :code:`X[i]` is one such point in
			:math:`\mathbf{R}^n`.
		c: array-like
			The vector to take the inner product with.

		Returns
		-------
		Vc: np.array (M,)
			Product of Vandermonde matrix and :math:`\mathbf c`
		"""
		X = X.reshape(-1, self.dim)
		X = self._scale(np.array(X))
		M = X.shape[0]
		c = np.array(c)
		assert len(self.indices) == c.shape[0]

		if len(c.shape) == 2:
			oneD = False
		else:
			c = c.reshape(-1,1)
			oneD = True

		V_coordinate = [self.vander(X[:,k], self.degree) for k in range(self.dim)]
		out = np.zeros((M, c.shape[1]))
		for j, alpha in enumerate(self.indices):

			# If we have a non-zero coefficient
			if np.max(np.abs(c[j,:])) > 0.:
				col = np.ones(M)
				for ell in range(self.dim):
					col *= V_coordinate[ell][:,alpha[ell]]

				for k in range(c.shape[1]):
					out[:,k] += c[j,k]*col
		if oneD:
			out = out.flatten()
		return out

	def DV(self, X):
		r""" Column-wise derivative of the Vandermonde matrix

		Given points :math:`\mathbf x_i \in \mathbb{R}^n`,
		this creates the Vandermonde-like matrix whose entries
		correspond to the derivatives of each of basis elements;
		i.e.,

		.. math::

			[\mathbf{V}]_{i,j} = \left. \frac{\partial}{\partial x_k} \psi_j(\mathbf{x})
				\right|_{\mathbf{x} = \mathbf{x}_i}.

		Parameters
		----------
		X: array-like (M, n)
			Points at which to evaluate the basis at where :code:`X[i]` is one such point in
			:math:`\mathbf{R}^n`.

		Returns
		-------
		Vp: np.array (M, N, n)
			Derivative of Vandermonde matrix where :code:`Vp[i,j,:]`
			is the gradient of :code:`V[i,j]`.
		"""
		X = X.reshape(-1, self.dim)
		X = self._scale(np.array(X))
		M = X.shape[0]
		V_coordinate = [self.vander(X[:,k], self.degree) for k in range(self.dim)]

		N = len(self.indices)
		DV = np.ones((M, N, self.dim), dtype = X.dtype)

		try:
			dscale = self._dscale()
		except NotImplementedError:
			dscale = np.ones(X.shape[1])


		for k in range(self.dim):
			for j, alpha in enumerate(self.indices):
				for q in range(self.dim):
					if q == k:
						DV[:,j,k] *= np.dot(V_coordinate[q][:,0:-1], self.Dmat[alpha[q],:])
					else:
						DV[:,j,k] *= V_coordinate[q][:,alpha[q]]
			# Correct for transform
			DV[:,:,k] *= dscale[k]

		return DV

	def DDV(self, X):
		r""" Column-wise second derivative of the Vandermonde matrix

		Given points :math:`\mathbf x_i \in \mathbb{R}^n`,
		this creates the Vandermonde-like matrix whose entries
		correspond to the derivatives of each of basis elements;
		i.e.,

		.. math::

			[\mathbf{V}]_{i,j} = \left. \frac{\partial^2}{\partial x_k\partial x_\ell} \psi_j(\mathbf{x})
				\right|_{\mathbf{x} = \mathbf{x}_i}.

		Parameters
		----------
		X: array-like (M, n)
			Points at which to evaluate the basis at where :code:`X[i]` is one such point in
			:math:`\mathbf{R}^m`.

		Returns
		-------
		Vpp: np.array (M, N, n, n)
			Second derivative of Vandermonde matrix where :code:`Vpp[i,j,:,:]`
			is the Hessian of :code:`V[i,j]`.
		"""
		X = X.reshape(-1, self.dim)
		X = self._scale(np.array(X))
		M = X.shape[0]
		V_coordinate = [self.vander(X[:,k], self.degree) for k in range(self.dim)]

		N = len(self.indices)
		DDV = np.ones((M, N, self.dim, self.dim), dtype = X.dtype)

		try:
			dscale = self._dscale()
		except NotImplementedError:
			dscale = np.ones(X.shape[1])


		for k in range(self.dim):
			for ell in range(k, self.dim):
				for j, alpha in enumerate(self.indices):
					for q in range(self.dim):
						if q == k == ell:
							# We need the second derivative
							eq = np.zeros(self.degree+1)
							eq[alpha[q]] = 1.
							der2 = self.polyder(eq, 2)
							DDV[:,j,k,ell] *= V_coordinate[q][:,0:len(der2)].dot(der2)
						elif q == k or q == ell:
							DDV[:,j,k,ell] *= np.dot(V_coordinate[q][:,0:-1], self.Dmat[alpha[q],:])
						else:
							DDV[:,j,k,ell] *= V_coordinate[q][:,alpha[q]]

				# Correct for transform
				DDV[:,:,k, ell] *= dscale[k]*dscale[ell]
				DDV[:,:,ell, k] = DDV[:,:,k, ell]
		return DDV

	def roots(self, coef):
		if self.dim > 1:
			raise NotImplementedError
		r = self.polyroots(coef)
		return r*(self._ub[0] - self._lb[0])/2.0 + (self._ub[0] + self._lb[0])/2.


class MonomialTensorBasis(PolynomialTensorBasis):
	"""A tensor product basis of bounded total degree built from the monomials"""
	def __init__(self, *args, **kwargs):
		self.vander = np.polynomial.polynomial.polyvander
		self.polyder = np.polynomial.polynomial.polyder
		self.polyroots = np.polynomial.polynomial.polyroots
		PolynomialTensorBasis.__init__(self, *args, **kwargs)


class LegendreTensorBasis(PolynomialTensorBasis):
	"""A tensor product basis of bounded total degree built from the Legendre polynomials

	"""
	def __init__(self, *args, **kwargs):
		self.vander = legvander
		self.polyder = legder
		self.polyroots = legroots
		PolynomialTensorBasis.__init__(self, *args, **kwargs)

class ChebyshevTensorBasis(PolynomialTensorBasis):
	"""A tensor product basis of bounded total degree built from the Chebyshev polynomials

	"""
	def __init__(self, *args, **kwargs):
		self.vander = chebvander
		self.polyder = chebder
		self.polyroots = chebroots
		PolynomialTensorBasis.__init__(self, *args, **kwargs)



class LaguerreTensorBasis(PolynomialTensorBasis):
	"""A tensor product basis of bounded total degree built from the Laguerre polynomials

	"""
	def __init__(self, *args, **kwargs):
		self.vander = lagvander
		self.polyder = lagder
		self.polyroots = lagroots
		PolynomialTensorBasis.__init__(self, *args, **kwargs)

class HermiteTensorBasis(PolynomialTensorBasis):
	"""A tensor product basis of bounded total degree built from the Hermite polynomials

	"""
	def __init__(self, *args, **kwargs):
		self.vander = hermvander
		self.polyder = hermder
		self.polyroots = hermroots
		PolynomialTensorBasis.__init__(self, *args, **kwargs)

	def _set_scale(self, X):
		self._mean = np.mean(X, axis = 0)
		self._std = np.std(X, axis = 0)

	def _scale(self, X):
		try:
			return (X - self._mean[None,:])/self._std[None,:]/np.sqrt(2)
		except AttributeError:
			return X

	def _dscale(self):
		try:
			return 1./self._std/np.sqrt(2)
		except AttributeError:
			raise NotImplementedError

	def roots(self, coef):
		if self.dim > 1:
			raise NotImplementedError
		r = hermroots(coef)
		return r*self._std[0]*np.sqrt(2) + self._mean[0]


class ArnoldiPolynomialBasis(Basis):
	r""" Construct a stable polynomial basis for arbitrary points using Vandermonde+Arnoldi
	"""
	def __init__(self, degree, X):
		self.X = np.copy(np.atleast_2d(X))
		self.dim = self.X.shape[1]
		self.degree = int(degree)
		self.indices = index_set(self.degree, self.dim)

		self.Q, self.R = self.arnoldi()

	def __len__(self):
		return len(self.indices)

	def _update_vec(self, ids):
		# Determine which column to multiply by
		diff = self.indices - ids
		# Here we pick the most recent column that is one off
		j = np.max(np.argwhere( (np.sum(np.abs(diff), axis = 1) <= 1) & (np.min(diff, axis = 1) == -1)))
		i = int(np.argwhere(diff[j] == -1))
		return i, j

	def arnoldi(self):
		r""" Apply the Arnoldi proceedure to build up columns of the Vandermonde matrix
		"""
		idx = self.indices
		M = self.X.shape[0]

		# Allocate memory for matrices
		Q = np.zeros((M, len(idx)))
		R = np.zeros((len(idx), len(idx)))

		# Generate columns of the Vandermonde matrix
		iteridx = enumerate(idx)
		# As the first column is the ones vector, we treat it as a special case
		next(iteridx)
		Q[:,0] = 1/np.sqrt(M)
		R[0,0] = np.sqrt(M)

		# Now work on the remaining columns
		for k, ids in iteridx:
			i, j = self._update_vec(ids)
			# Form new column
			q = self.X[:,i] * Q[:,j]

			for j in range(k):
				R[j,k] = Q[:,j].T @ q
				q -= R[j,k]*Q[:,j]

			R[k,k] = np.linalg.norm(q)
			Q[:,k] = q/R[k,k]

		return Q, R

	def arnoldi_X(self, X):
		r""" Generate a Vandermonde matrix corresponding to a different set of points
		"""
		W = np.zeros((X.shape[0], len(self.indices)), dtype = X.dtype)

		iteridx = enumerate(self.indices)
		# As the first column is the ones vector, we treat it as a special case
		next(iteridx)
		W[:,0] = 1/self.R[0,0]

		# Now work on the remaining columns
		for k, ids in iteridx:
			i, j = self._update_vec(ids)
			# Form new column
			w = X[:,i] * W[:,j]

			for j in range(k):
				w -= self.R[j,k]*W[:,j]

			W[:,k] = w/self.R[k,k]

		return W


	def V(self, X = None):
		if X is None or np.array_equal(X, self.X):
			return self.Q
		else:
			return self.arnoldi_X(X)


	def DV(self, X = None):
		if X is None or np.array_equal(X, self.X):
			X = self.X
			V = self.Q
		else:
			V = self.arnoldi_X(X)

		M = X.shape[0]
		N = self.Q.shape[1]
		n = self.X.shape[1]
		DV = np.zeros((M, N, n), dtype = self.Q.dtype)

		for ell in range(n):
			index_iterator = enumerate(self.indices)
			next(index_iterator)
			for k, ids in index_iterator:
				i, j = self._update_vec(ids)
				# Q[:,k] = X[:,i] * Q[:,j] - sum_s Q[:,s] * R[s, k]
				if i == ell:
					DV[:,k,ell] = V[:,j] + X[:,i] * DV[:,j,ell] - DV[:,0:k,ell] @ self.R[0:k,k]
				else:
					DV[:,k,ell] = X[:,i] * DV[:,j,ell] - DV[:,0:k,ell] @ self.R[0:k,k]
				DV[:,k,ell] /= self.R[k,k]

		return DV

	def DDV(self, X = None):
		raise NotImplementedError



from __future__ import print_function, division
""" misc. utilities and defintions
"""
import numpy as np

def merge(x, y):
	z = x.copy()
	z.update(y)
	return z


def check_sample_inputs(X, fX, grads):
	if X is not None and fX is not None:
		X = np.array(X)
		fX = np.array(fX).flatten()
		assert len(X) == len(fX), "Number of samples doesn't match number of evaluations"
	else:
		X = None
		fX = None

	if grads is not None:
		grads = np.array(grads)
		if X is not None:
			assert X.shape[1] == grads.shape[1], "Dimensions of gradients doesn't match dimension of samples"

	if X is None:
		X = np.zeros((0, grads.shape[1]))
		fX = np.zeros((0,))
	if grads is None:
		grads = np.zeros((0, X.shape[1]))
	return X, fX, grads	



import unittest
import pandas as pd

from context import grama as gr
DF = gr.Intention()

class TestDataHelpers(unittest.TestCase):
    def setUp(self):
        pass

    def test_df_equal(self):
        df1 = pd.DataFrame(dict(x=[0], y=[0]))
        df2 = pd.DataFrame(dict(x=[0]))

        self.assertTrue(gr.df_equal(df1, df1))
        self.assertTrue(gr.df_equal(df1, df2) == False)

    def test_df_make(self):
        # Check correctness
        df_true = pd.DataFrame(dict(x=[0, 1], y=[0, 0], z=[1, 1]))
        df_res = gr.df_make(x=[0, 1], y=[0], z=1)

        self.assertTrue(gr.df_equal(df_true, df_res))

        # Check for mismatch
        with self.assertRaises(ValueError):
            gr.df_make(x=[1, 2, 3], y=[1, 2])

        # Catch an intention operator
        with self.assertRaises(ValueError):
            gr.df_make(y=DF.x)

    def test_df_grid(self):
        # Check correctness
        df_res = gr.df_grid(
            x=[0, 1, 2],
            y=["A", "B"],
            z=1,
            w="x",
        )
        df_true = pd.DataFrame(dict(
            x=[  0,   1,   2,   0,   1,   2],
            y=["A", "A", "A", "B", "B", "B"],
            z=[  1,   1,   1,   1,   1,   1],
            w=["x", "x", "x", "x", "x", "x"],
        ))

        self.assertTrue(gr.df_equal(df_true, df_res))

        # Catch an intention operator
        with self.assertRaises(ValueError):
            gr.df_grid(y=DF.x)



__all__ = [
    "df_equal",
    "df_make",
    "df_grid",
]

from pandas import DataFrame
from pandas.testing import assert_frame_equal

from .dfply import Intention
from .tools import tran_outer

## Safe length-checker
def safelen(x):
    try:
        if isinstance(x, str):
            raise TypeError
        return len(x)

    except TypeError:
        return 1


## DataFrame constructor utility
def df_make(**kwargs):
    r"""Construct a DataFrame

    Helper function to construct a DataFrame. A common use-case is to use df_make() to pass values to the df (and related) keyword arguments succinctly.

    Kwargs:
        varname (iterable): Column for constructed dataframe; column name inferred from variable name.

    Returns:
        DataFrame: Constructed DataFrame

    Preconditions:
        All provided iterables must have identical length or be of length one.
        All provided variable names (keyword arguments) must be distinct.

    Examples::

        import grama as gr
        from models import make_test
        md = make_test()
        (
            md
            >> gr.ev_sample(
                n=1e3,
                df_det=gr.df_make(x2=[1, 2])
            )
        )

    """
    ## Catch passed Intention operator
    if any([isinstance(v, Intention) for v in kwargs.values()]):
        raise ValueError(
            "df_make() does not support the Intention operator; " +
            "did you mean to use a DataFrame argument?\n\n" +
            "A common mistake is to write\n\n" +
            "    lambda df: gr.df_make(y=DF.x) # Incorrect\n\n" +
            "rather than\n\n" +
            "    lambda df: gr.df_make(y=df.x) # Correct"
        )

    ## Check lengths
    lengths = [safelen(v) for v in kwargs.values()]
    length_max = max(lengths)

    if not all([(l == length_max) | (l == 1) for l in lengths]):
        raise ValueError("Column lengths must be identical or one.")

    ## Construct dataframe
    df_res = DataFrame()
    for key in kwargs.keys():
        try:
            if len(kwargs[key]) > 1:
                df_res[key] = kwargs[key]
            else:
                df_res[key] = [kwargs[key][0]] * length_max
        except TypeError:
            df_res[key] = [kwargs[key]] * length_max

    return df_res


## DataFrame equality checker
def df_equal(df1, df2, close=False, precision=3):
    """Check DataFrame equality

    Check that two dataframes have the same columns and values. Allows column order to differ.

    Args:
        df1 (DataFrame): Comparison input 1
        df2 (DataFrame): Comparison input 2

    Returns:
        bool: Result of comparison

    """

    if not set(df1.columns) == set(df2.columns):
        return False

    if close:
        try:
            assert_frame_equal(
                df1[df2.columns],
                df2,
                check_dtype=False,
                check_exact=False,
                rtol=1e5
            )
            return True
        except:
            return False
    else:
        return df1[df2.columns].equals(df2)

## DataFrame constructor utility; outer product
def df_grid(**kwargs):
    r"""Construct a DataFrame as outer-product

    Helper function to construct a DataFrame as an outer-product of the given columns.

    Kwargs:
        varname (iterable): Column for constructed dataframe; column name inferred from variable name.

    Returns:
        DataFrame: Constructed DataFrame

    Preconditions:
        All provided variable names (keyword arguments) must be distinct.

    Examples::

        import grama as gr
        ## Make an empty DataFrame
        gr.df_grid()
        ## Create a row for every pair of values (6 rows total)
        gr.df_grid(x=["A", "B"], y=[1, 2, 3])

    """
    ## Catch passed Intention operator
    if any([isinstance(v, Intention) for v in kwargs.values()]):
        raise ValueError(
            "df_grid() does not support the Intention operator; " +
            "did you mean to use a DataFrame argument?"
        )

    ## Construct dataframe
    df_res = DataFrame()
    for key in kwargs.keys():
        # Switch based on argument length
        l = safelen(kwargs[key])
        if l == 1:
            df_tmp = DataFrame(columns=[key], data=[kwargs[key]])
        else:
            df_tmp = DataFrame(columns=[key], data=kwargs[key])

        df_res = tran_outer(df_res, df_outer=df_tmp)

    return df_res



import numpy as np
import pandas as pd
from scipy.stats import norm
import unittest

from context import grama as gr
from context import data

X = gr.Intention()

##==============================================================================
## filter helper tests
##==============================================================================


class TestSubset(unittest.TestCase):
    def test_dropna(self):
        df = gr.df_make(x=[1.0, 2.0, 3.0], y=[1.0, np.nan, 3.0], z=[1.0, 2.0, np.nan])

        df_true_default = gr.df_make(x=[1.0], y=[1.0], z=[1.0])
        df_true_y = gr.df_make(x=[1.0, 3.0], y=[1.0, 3.0], z=[1.0, np.nan])

        df_res_default = df >> gr.tf_dropna()
        df_res_y = df >> gr.tf_dropna(subset=["y"])

        self.assertTrue(gr.df_equal(df_true_default, df_res_default))
        self.assertTrue(gr.df_equal(df_true_y, df_res_y))



import numpy as np
import pandas as pd
import unittest

from context import grama as gr
from context import data

DF = gr.Intention()

##==============================================================================
## summarization test functions
##==============================================================================

class TestSummarize(unittest.TestCase):
    def test_summarize(self):
        df = gr.df_make(
            x=["A", "A", "B", "B"],
            y=["A", "B", "A", "B"],
        )
        df_true1 = gr.df_make(
            x=["A", "B"],
            n=[ 2,    2],
        )
        df_true2 = gr.df_make(
            x=["A", "A", "B", "B"],
            y=["A", "B", "A", "B"],
            n=[  1,   1,   1,   1],
        )

        df_res1 = (
            df
            >> gr.tf_count(DF.x)

        )
        self.assertTrue(df_true1.equals(df_res1))

        df_res2 = (
            df
            >> gr.tf_count(DF.x, DF.y)
        )
        self.assertTrue(df_true2.equals(df_res2))



from __future__ import division

import numpy as np
from scipy.spatial.distance import pdist

from .domain import TOL
from .linineq import LinIneqDomain
from ..exceptions import EmptyDomainException


class BoxDomain(LinIneqDomain):
	r""" Implements a domain specified by box constraints

	Given a set of lower and upper bounds, this class defines the domain

	.. math::

		\mathcal{D} := \lbrace \mathbf{x} \in \mathbb{R}^m : \text{lb} \le \mathbf{x} \le \text{ub} \rbrace \subset \mathbb{R}^m.

	Parameters
	----------
	lb: array-like (m,)
		Lower bounds
	ub: array-like (m,)
		Upper bounds
	"""
	def __init__(self, lb, ub, names = None):
		LinIneqDomain.__init__(self, lb = lb, ub = ub, names = names)	
		#assert np.all(np.isfinite(lb)) and np.all(np.isfinite(ub)), "Both lb and ub must be finite to construct a box domain"

	@property
	def is_empty(self):
		try:
			return self._empty
		except AttributeError:
			self._empty = np.any(self.lb > self.ub)
			self._point = False
			self._unbounded = False
			return self._empty
	
	@property
	def is_point(self):
		try:
			return self._point
		except AttributeError:
			self._point = np.all(np.abs(self.ub - self.lb) < TOL)
			return self._point

	@property
	def is_unbounded(self):
		try:
			return self._unbounded
		except AttributeError:
			self._unbounded = np.any(np.isinf(self.lb)) | np.any(np.isinf(self.ub))
			return self._unbounded


	# Due to the simplicity of this domain, we can use a more efficient sampling routine
	def _sample(self, draw = 1):
		if self.is_empty:
			raise EmptyDomainException
		x_sample = np.random.uniform(self.lb, self.ub, size = (draw, len(self)))
		return x_sample

	def _corner(self, p, **kwargs):
		# Since the domain is a box, we can find the corners simply by looking at the sign of p
		x = np.copy(self.lb)
		I = (p>=0)
		x[I] = self.ub[I]
		if not self.isinside(x):
			raise EmptyDomainException
		return x

	def _extent(self, x, p):
		return self._extent_bounds(x, p)

	def _isinside(self, X, tol = TOL):
		return self._isinside_bounds(X, tol = tol)

	def _normalized_domain(self, **kwargs):
		names_norm = [name + ' (normalized)' for name in self.names]
		return BoxDomain(lb = self.lb_norm, ub = self.ub_norm, names = names_norm)

	@property
	def A(self): return np.zeros((0,len(self)))
	
	@property
	def b(self): return np.zeros((0))
	
	@property
	def A_eq(self): return np.zeros((0,len(self)))
	
	@property
	def b_eq(self): return np.zeros((0))



# Subspace based dimension reduction techniques
from __future__ import division, print_function

import os
# Necessary for running in headless enviornoments
if 'DISPLAY' not in os.environ:
	import matplotlib
	matplotlib.use("Agg")
import matplotlib.pyplot as plt


import numpy as np
import scipy.linalg
import scipy.optimize
import scipy.sparse


from .pgf import PGF
from .misc import merge

__all__ = ['SubspaceBasedDimensionReduction',
	'ActiveSubspace',
	]

class SubspaceBasedDimensionReduction(object):
	r""" Abstract base class for Subspace-Based Dimension Reduction

	Given a function :math:`f : \mathcal{D} \to \mathbb{R}`,
	subspace-based dimension reduction identifies a subspace,
	described by a matrix :math:`\mathbf{U} \in \mathbb{R}^{m\times n}`
	with orthonormal columns for some :math:`n \le m`.

	"""
	@property
	def U(self):
		""" A matrix defining the 'important' directions

		Returns
		-------
		np.ndarray (m, n):
			Matrix with orthonormal columns defining the important directions in decreasing order
			of precidence.
		"""
		raise NotImplementedError


	def shadow_plot(self, X = None, fX = None, dim = 1, U = None, ax = 'auto', pgfname = None):
		r""" Draw a shadow plot


		Parameters
		----------
		X: array-like (N,m)
			Input coordinates for function samples
		fX: array-like (N,)
			Values of function at sample points
		dim: int, [1,2]
			Dimension of shadow plot
		U: array-like (?,m); optional
			Subspace onto which to project the data; defaults to the subspace identifed by this class
		ax: 'auto', matplotlib.pyplot.axis, or None
			Axis on which to draw the shadow plot

		Returns
		-------
		ax: matplotlib.pyplot.axis
			Axis on which the plot is drawn
		"""
		if ax == 'auto':
			if dim == 1:
				fig, ax = plt.subplots(figsize = (6,6))
			else:
				# Hack so that plot is approximately square after adding colorbar
				fig, ax = plt.subplots(figsize = (7.5,6))

		if X is None:
			X = self.X

		# Check dimensions
		X = np.atleast_2d(X)
		assert X.shape[1] == len(self), "Samples do not match dimension of space"

		if U is None:
			U = self.U
		else:
			if len(U.shape) == 1:
				U = U.reshape(len(self),1)
			else:
				assert U.shape[0] == len(self), "Dimensions do not match"


		if dim == 1:
			if ax is not None:
				ax.plot(X.dot(U[:,0]), fX, 'k.')
				ax.set_xlabel(r'active coordinate $\mathbf{u}^\top \mathbf{x}$')
				ax.set_ylabel(r'$f(\mathbf{x})$')

			if pgfname is not None:
				pgf = PGF()
				pgf.add('y', X.dot(U[:,0]))
				pgf.add('fX', fX)
				pgf.write(pgfname)

		elif dim == 2:
			Y = U[:,0:2].T.dot(X.T).T

			if ax is not None:
				sc = ax.scatter(Y[:,0], Y[:,1], c = fX.flatten(), s = 3)
				ax.set_xlabel(r'active coordinate 1 $\mathbf{u}_1^\top \mathbf{x}$')
				ax.set_ylabel(r'active coordinate 2 $\mathbf{u}_2^\top \mathbf{x}$')

				plt.colorbar(sc).set_label('f(x)')

			if pgfname is not None:
				pgf = PGF()
				pgf.add('y1', Y[:,0])
				pgf.add('y2', Y[:,1])
				pgf.add('fX', fX.flatten())
				pgf.write(pgfname)

		else:
			raise NotImplementedError

		return ax



	def _init_dim(self, X = None, grads = None):
		if X is not None and len(X) > 0:
			self._dimension = len(X[0])
		elif grads is not None:
			self._dimension = len(grads[0])
		else:
			raise Exception("Could not determine dimension of ambient space")


	def __len__(self):
		return self._dimension

	@property
	def X(self):
		return np.zeros((0,len(self)))

	@property
	def fX(self):
		return np.zeros((0,len(self)))

	@property
	def grads(self):
		return np.zeros((0,len(self)))

	def _fix_subspace_signs(self, U, X = None, fX = None, grads = None):
		r""" Orient the subspace so that the average slope is positive

		Since subspaces have no associated direction (they are invariant to a sign flip)
		here we fix the sign such that the function is increasing on average along the direction
		u_i.  This approach uses either gradient or sample information, with a preference for
		gradient information if it is availible.
		"""
		if grads is not None and len(grads) > 0:
			return self._fix_subspace_signs_grads(U, grads)
		else:
			return self._fix_subspace_signs_samps(U, X, fX)

	def _fix_subspace_signs_samps(self, U, X, fX):
		sgn = np.zeros(len(U[0]))
		for k in range(len(U[0])):
			for i in range(len(X)):
				for j in range(i+1, len(X)):
					denom = U[:,k] @ (X[i] - X[j])
					if np.abs(denom) > 0:
						sgn[k] += (fX[i] - fX[j])/denom

		# If the sign is zero, keep the current orientation
		sgn[sgn == 0] = 1
		return U.dot(np.diag(np.sign(sgn)))

	def _fix_subspace_signs_grads(self, U, grads):
		return U.dot(np.diag(np.sign(np.mean(grads.dot(U), axis = 0))))


	def approximate_lipschitz(self, X = None, fX = None, grads = None,  dim = None):
		r""" Approximate the Lipschitz matrix on the low-dimensional subspace
		"""
		raise NotImplementedError


class ActiveSubspace(SubspaceBasedDimensionReduction):
	r"""Computes the active subspace gradient samples

	Given the function :math:`f:\mathcal{D} \to \mathbb{R}`,
	the active subspace is defined as the eigenvectors corresponding to the
	largest eigenvalues of the average outer-product of gradients:

	.. math::

		\mathbf{C} := \int_{\mathbf{x}\in \mathcal{D}} \nabla f(\mathbf{x}) \nabla f(\mathbf{x})^\top \  \mathrm{d}\mathbf{x}
		\in \mathbb{R}^{m\times m}.

	By default, this class assumes that we are provided with gradients
	evaluated at random samples over the domain and estimates the matrix :math:`\mathbf{C}`
	using Monte-Carlo integration. However, if provided a weight corresponding to a quadrature rule,
	this will be used instead to approximate this matrix; i.e.,

	.. math::

		\mathbf{C} \approx \sum_{i=1}^N w_i \nabla f(\mathbf{x}_i) \nabla f(\mathbf{x}_i)^\top.

	"""
	def __init__(self):
		self._U = None
		self._s = None


	def fit(self, grads, weights = None):
		r""" Find the active subspace

		Parameters
		----------
		grads: array-like (N,m)
			Gradient samples of function (tacitly assumed to be uniform on the domain
			or from a quadrature rule with corresponding weight).
		weights: array-like (N,), optional
			Weights corresponding to a quadrature rule associated with the samples of the gradient.

		"""
		self._init_dim(grads = grads)

		self._grads = np.array(grads).reshape(-1,len(self))
		N = len(self._grads)
		if weights is None:
			weights = np.ones(N)/N

		self._weights = np.array(weights)
		self._U, self._s, VT = scipy.linalg.svd(np.sqrt(self._weights)*self._grads.T)
		# Pad s with zeros if we don't have as many gradient samples as dimension of the space
		self._s = np.hstack([self._s, np.zeros(self._dimension - len(self._s))])
		self._C = self._U @ np.diag(self._s**2) @ self._U.T

		# Fix +/- scaling so average gradient is positive
		self._U = self._fix_subspace_signs_grads(self._U, self._grads)


	def fit_function(self, fun, N_gradients):
		r""" Automatically estimate active subspace using a quadrature rule

		Parameters
		----------
		fun: Function
			function object for which to estimate the active subspace via the average outer-product of gradients
		N_gradients: int
			Maximum number of gradient samples to use
		"""
		X, w = fun.domain.quadrature_rule(N_gradients)
		grads = fun.grad(X)
		self.fit(grads, w)


	@property
	def U(self):
		return np.copy(self._U)

	@property
	def C(self):
		return self._C

	@property
	def singvals(self):
		return self._s

	# TODO: Plot of eigenvalues (with optional boostrapped estimate)

	# TODO: Plot of eigenvector angles with bootstrapped replicates.



import numpy as np
import pandas as pd
import unittest

from context import grama as gr
from context import data

X = gr.Intention()

##==============================================================================
## summarization test functions
##==============================================================================


class TestSummarize(unittest.TestCase):
    def test_summarize(self):
        p = pd.DataFrame(
            {
                "price_mean": [data.df_diamonds.price.mean()],
                "price_std": [data.df_diamonds.price.std()],
            }
        )
        self.assertTrue(
            p.equals(
                data.df_diamonds
                >> gr.tf_summarize(price_mean=X.price.mean(), price_std=X.price.std())
            )
        )

        pcut = pd.DataFrame({"cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"]})
        pcut["price_mean"] = [
            data.df_diamonds[data.df_diamonds.cut == c].price.mean()
            for c in pcut.cut.values
        ]
        pcut["price_std"] = [
            data.df_diamonds[data.df_diamonds.cut == c].price.std()
            for c in pcut.cut.values
        ]
        self.assertTrue(
            pcut.equals(
                data.df_diamonds
                >> gr.tf_group_by("cut")
                >> gr.tf_summarize(price_mean=X.price.mean(), price_std=X.price.std())
            )
        )

    def test_summarize_each(self):
        to_match = pd.DataFrame(
            {
                "price_mean": [np.mean(data.df_diamonds.price)],
                "price_var": [np.var(data.df_diamonds.price)],
                "depth_mean": [np.mean(data.df_diamonds.depth)],
                "depth_var": [np.var(data.df_diamonds.depth)],
            }
        )
        to_match = to_match[["price_mean", "price_var", "depth_mean", "depth_var"]]

        test1 = data.df_diamonds >> gr.tf_summarize_each([np.mean, np.var], X.price, 4)
        test2 = data.df_diamonds >> gr.tf_summarize_each(
            [np.mean, np.var], X.price, "depth"
        )
        self.assertTrue(to_match.equals(test1))
        self.assertTrue(to_match.equals(test2))

        group = pd.DataFrame({"cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"]})
        group["price_mean"] = [
            np.mean(data.df_diamonds[data.df_diamonds.cut == c].price)
            for c in group.cut.values
        ]
        group["price_var"] = [
            np.var(data.df_diamonds[data.df_diamonds.cut == c].price)
            for c in group.cut.values
        ]
        group["depth_mean"] = [
            np.mean(data.df_diamonds[data.df_diamonds.cut == c].depth)
            for c in group.cut.values
        ]
        group["depth_var"] = [
            np.var(data.df_diamonds[data.df_diamonds.cut == c].depth)
            for c in group.cut.values
        ]

        group = group[["cut", "price_mean", "price_var", "depth_mean", "depth_var"]]

        test1 = (
            data.df_diamonds
            >> gr.tf_group_by(X.cut)
            >> gr.tf_summarize_each([np.mean, np.var], X.price, 4)
        )
        test2 = (
            data.df_diamonds
            >> gr.tf_group_by("cut")
            >> gr.tf_summarize_each([np.mean, np.var], X.price, "depth")
        )

        self.assertTrue(group.equals(test1))
        self.assertTrue(group.equals(test2))



__all__ = ["make_piston", "make_piston_rand"]

from grama import cp_bounds, cp_copula_independence, cp_vec_function, cp_marginals, \
    Model, df_make, marg_mom
from numpy import sqrt, array, Inf, float64, pi


def make_piston():
    """Piston cycle time

    A simple model for the cycle time for a reciprocating piston within a cylinder. All inputs are treated as deterministic variables.

    Deterministic Variables:
        m: Piston weight (kg)
        s: Piston surface area (m^2)
        v_0: Initial gas volume (m^3)
        k: Spring coefficient (N/m)
        p_0: Atmospheric pressure (N/m^2)
        t_a: Ambient temperature (K)
        t_0: Filling gas temperature (K)

    Random Variables:
        (None)

    Outputs:
        t_cyc: cycle time (s)

    References:
        Kenett, R., & Zacks, S. (1998). Modern industrial statistics: design and control of quality and reliability. Pacific Grove, CA: Duxbury press.
        Moon, H. (2010). Design and Analysis of Computer Experiments for Screening Input Variables (Doctoral dissertation, Ohio State University).

    """
    md = (
        Model(name = "Piston cycle time")
        >> cp_vec_function(
            fun=lambda df: df_make(
                a=df.p_0 * df.s
                 +19.62 * df.m
                 -df.k * df.v_0 / df.s
            ),
            var=["p_0", "s", "m", "k", "v_0"],
            out=["a"],
            name="Intermediate calculation 1",
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                v=0.5*df.s/df.k * (
                    sqrt(df.a**2 + 4*df.k*df.p_0*df.v_0/df.t_0*df.t_a) - df.a
                )
            ),
            var=["s", "k", "a", "k", "p_0", "v_0", "t_0", "t_a"],
            out=["v"],
            name="Intermediate calculation 2",
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                t_cyc=2*pi*sqrt(
                    df.m / (df.k + df.s**2 * df.p_0*df.v_0/df.t_0 * df.t_a/df.v**2)
                )
            ),
            var=["m", "k", "s", "p_0", "v_0", "t_0", "t_a", "v"],
            out=["t_cyc"],
            name="Cycle time",
        )
        >> cp_bounds(
            m=(30, 60),          # kg
            s=(0.005, 0.020),    # m^2
            v_0=(0.002, 0.010),  # m^3
            k=(1000, 5000),      # N/m
            p_0=(90000, 110000), # Pa
            t_a=(290, 296),      # K
            t_0=(340, 360),      # K
        )
    )

    return md

def make_piston_rand():
    """Piston cycle time

    A simple model for the cycle time for a reciprocating piston within a cylinder. Some inputs are treated as deterministic nominal conditions with additive randomness.

    Deterministic Variables:
        m: Piston weight (kg)
        s: Piston surface area (m^2)
        v_0: Initial gas volume (m^3)

        k:   (Nominal) Spring coefficient (N/m)
        p_0: (Nominal) Atmospheric pressure (N/m^2)
        t_a: (Nominal) Ambient temperature (K)
        t_0: (Nominal) Filling gas temperature (K)

    Random Variables:
        dk:   Fluctuation in k   (unitless)
        dp_0: Fluctuation in p_0 (unitless)
        dt_a: Fluctuation in t_a (unitless)
        dt_0: Fluctuation in t_0 (unitless)

    Outputs:
        t_cyc: cycle time (s)

    References:
        Kenett, R., & Zacks, S. (1998). Modern industrial statistics: design and control of quality and reliability. Pacific Grove, CA: Duxbury press.
        Moon, H. (2010). Design and Analysis of Computer Experiments for Screening Input Variables (Doctoral dissertation, Ohio State University).

    """
    md = (
        Model(name = "Piston cycle time, with variability")
        >> cp_vec_function(
            fun=lambda df: df_make(
                rk=df.k * (1 + df.dk),
                rp_0=df.p_0 * (1 + df.dp_0),
                rt_a=df.t_a * (1 + df.dt_a),
                rt_0=df.t_0 * (1 + df.dt_0),
            ),
            var=["k", "p_0", "t_a", "t_0",
                 "dk", "dp_0", "dt_a", "dt_0"],
            out=["rk", "rp_0", "rt_a", "rt_0"],
            name="Random operating disturbances",
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                a=df.rp_0 * df.s
                 +19.62 * df.m
                 -df.rk * df.v_0 / df.s
            ),
            var=["rp_0", "s", "m", "rk", "v_0"],
            out=["a"],
            name="Intermediate calculation 1",
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                v=0.5*df.s/df.rk * (
                    sqrt(df.a**2 + 4*df.rk*df.rp_0*df.v_0/df.rt_0*df.rt_a) - df.a
                )
            ),
            var=["s", "a", "rk", "rp_0", "v_0", "rt_0", "rt_a"],
            out=["v"],
            name="Intermediate calculation 2",
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                t_cyc=2*pi*sqrt(
                    df.m / (df.rk + df.s**2 * df.rp_0*df.v_0/df.rt_0 * df.rt_a/df.v**2)
                )
            ),
            var=["m", "rk", "s", "rp_0", "v_0", "rt_0", "rt_a", "v"],
            out=["t_cyc"],
            name="Cycle time",
        )
        >> cp_bounds(
            m=(30, 60),          # kg
            s=(0.005, 0.020),    # m^2
            v_0=(0.002, 0.010),  # m^3
            k=(1000, 5000),      # N/m
            p_0=(90000, 110000), # Pa
            t_a=(290, 296),      # K
            t_0=(340, 360),      # K
        )
        >> cp_marginals(
            dk=marg_mom("uniform", mean=0, sd=0.40),
            dp_0=marg_mom("uniform", mean=0, sd=0.40),
            dt_a=marg_mom("uniform", mean=0, sd=0.40),
            dt_0=marg_mom("uniform", mean=0, sd=0.40),
        )
        >> cp_copula_independence()
    )

    return md



import numpy as np
import pandas as pd
from scipy.stats import norm
import unittest
import networkx as nx

from context import grama as gr
from context import models


class TestFORM(unittest.TestCase):
    """Test implementations of FORM
    """

    def setUp(self):
        ## Linear limit state w/ MPP off initial guess
        self.beta_true = 3
        self.md = (
            gr.Model()
            >> gr.cp_function(
                fun=lambda x0, x1: self.beta_true * 2 - x0 - np.sqrt(3) * x1,
                var=2,
                out=["g"],
            )
            >> gr.cp_marginals(
                x0=dict(dist="norm", loc=0, scale=1, sign=1),
                x1=dict(dist="norm", loc=0, scale=1, sign=1),
            )
            >> gr.cp_copula_independence()
        )

        ## Linear limit state w/ lognormal marginals
        self.md_log = (
            gr.Model()
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(
                    g=gr.exp(gr.sqrt(2) * 1) - df.x * df.y
                ),
                var=["x", "y"],
                out=["g"]
            )
            >> gr.cp_marginals(
                x=dict(dist="lognorm", loc=0, scale=1, s=1),
                y=dict(dist="lognorm", loc=0, scale=1, s=1),
            )
            >> gr.cp_copula_independence()
        )
        self.df_mpp = gr.df_make(
            x=gr.exp(gr.sqrt(2)/2),
            y=gr.exp(gr.sqrt(2)/2),
            beta_g=1.0,
            g=0.0,
        )

        ## Cantilever beam for flatten test
        self.md_beam = models.make_cantilever_beam()

    def test_ria(self):
        ## Test accuracy
        df_res = self.md >> gr.ev_form_ria(df_det="nom", limits=["g"])
        self.assertTrue(np.allclose(df_res["beta_g"], [self.beta_true], atol=1e-3))

        ## Test MPP mapped correctly
        df_mpp = self.md_log >> gr.ev_form_ria(df_det="nom", limits=["g"])
        self.assertTrue(
            gr.df_equal(df_mpp, self.df_mpp[["x", "y", "beta_g"]], close=True)
        )

        ## Test flatten
        df_beam = self.md_beam >> gr.ev_form_ria(
            df_det="nom", limits=["g_stress", "g_disp"], append=False
        )
        self.assertTrue(df_beam.shape[0] == 1)

        ## Test other outputs
        limits = ["g_stress", "g_disp"]
        beta_names = ["beta_g_stress", "beta_g_disp"]
        # Return reliabilities
        df_rels = self.md_beam >> gr.ev_form_ria(
            df_det="nom", limits=limits, append=False, format="rels",
        )
        self.assertTrue(gr.df_equal(
            df_beam.apply(lambda col: norm.cdf(col) if col.name in beta_names else col)
                   .rename(columns={"beta_" + s: "rel_" + s for s in limits}),
            df_rels,
            close=True,
        ))
        # Return POFs
        df_pofs = self.md_beam >> gr.ev_form_ria(
            df_det="nom", limits=limits, append=False, format="pofs",
        )
        self.assertTrue(gr.df_equal(
            df_beam.apply(lambda col: 1-norm.cdf(col) if col.name in beta_names else col)
                   .rename(columns={"beta_" + s: "pof_" + s for s in limits}),
            df_pofs,
            close=True,
        ))

    def test_pma(self):
        ## Test accuracy
        df_res = self.md >> gr.ev_form_pma(df_det="nom", betas=dict(g=self.beta_true))
        self.assertTrue(np.allclose(df_res["g"], [0], atol=1e-3))

        ## Test MPP mapped correctly
        df_mpp = self.md_log >> gr.ev_form_pma(df_det="nom", betas=dict(g=1.0))
        self.assertTrue(
            gr.df_equal(df_mpp, self.df_mpp[["x", "y", "g"]], close=True)
        )

        ## Test flatten
        df_beam = self.md_beam >> gr.ev_form_pma(
            df_det="nom", betas={"g_stress": 3, "g_disp": 3}, append=False,
        )
        self.assertTrue(df_beam.shape[0] == 1)

        ## Specify reliabilities
        df_rels = self.md_beam >> gr.ev_form_pma(
            df_det="nom", rels={"g_stress": norm.cdf(3), "g_disp": norm.cdf(3)}, append=False,
        )
        self.assertTrue(gr.df_equal(df_beam, df_rels, close=True))

        ## Specify POFs
        df_pofs = self.md_beam >> gr.ev_form_pma(
            df_det="nom", pofs={"g_stress": 1 - norm.cdf(3), "g_disp": 1 - norm.cdf(3)}, append=False,
        )
        self.assertTrue(gr.df_equal(df_beam, df_pofs, close=True))



import numpy as np
import pandas as pd
import unittest

from context import grama as gr

##################################################
class TestContour(unittest.TestCase):
    def setUp(self):
        pass

    def test_contour(self):
        md1 = (
            gr.Model()
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(
                    f=df.x**2 + df.y**2,
                    g=df.x + df.y,
                ),
                var=["x", "y"],
                out=["f", "g"],
            )
            >> gr.cp_bounds(
                x=(-1, +1),
                y=(-1, +1),
            )
        )

        md2 = (
            gr.Model()
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(
                    f=(df.c) * df.x + (1 - df.c) * df.y,
                ),
                var=["x", "y", "c"],
                out=["f"],
            )
            >> gr.cp_bounds(
                x=(-1, +1),
                y=(-1, +1),
                c=(+0, +1),
            )
        )

        ## Basic functionality
        df_res1 = (
            md1
            >> gr.ev_contour(
                var=["x", "y"],
                out=["f", "g"],
                n_side=10, # Coarse, for speed
            )
        )
        # Contains correct columns
        self.assertTrue("x" in df_res1.columns)
        self.assertTrue("y" in df_res1.columns)

        df_res2 = (
            md2
            >> gr.ev_contour(
                var=["x", "y"],
                out=["f"],
                df=gr.df_make(c=[0, 1]),
                n_side=10, # Coarse, for speed
            )
        )
        # Contains auxiliary variable
        self.assertTrue("c" in df_res2.columns)

        df_res3 = (
            md1
            >> gr.ev_contour(
                var=["x", "y"],
                out=["g"],
                levels=dict(g=[-1, 0, +1]),
                n_side=10, # Coarse, for speed
            )
        )
        # Correct manual levels
        self.assertTrue(set(df_res3.level) == {-1, 0, +1})

        # Drops redundant (swept) inputs under-the-hood
        (
            md2
            >> gr.ev_contour(
                var=["x", "y"],
                out=["f"],
                df=gr.eval_nominal(md2, df_det="nom"),
                n_side=10, # Coarse, for speed
            )
        )

        # Correct manual levels
        with self.assertWarns(Warning):
            df_res4 = (
                md1
                >> gr.ev_contour(
                    var=["x", "y"],
                    out=["g"],
                    levels=dict(g=[-1, 0, +1, +1e3]),
                    n_side=10, # Coarse, for speed
                )
            )
            # Correct manual levels
            self.assertTrue(set(df_res4.level) == {-1, 0, +1})

        ## Check assertions
        # No var
        with self.assertRaises(ValueError):
            res = (
                md1
                >> gr.ev_contour(out=["f"])
            )

        # Incorrect number of inputs
        with self.assertRaises(ValueError):
            res = (
                md1
                >> gr.ev_contour(var=["x", "y", "z"])
            )

        # Unavailable inputs
        with self.assertRaises(ValueError):
            res = (
                md1
                >> gr.ev_contour(var=["foo", "bar"])
            )

        # Unsupported input
        with self.assertRaises(ValueError):
            res = (
                md2
                >> gr.ev_contour(
                    var=["x", "y"],
                    out=["f"],
                )
            )

        with self.assertRaises(ValueError):
            res = (
                md2
                >> gr.ev_contour(
                    var=["x", "y"],
                    out=["f"],
                    df=gr.df_make(foo=1)
                )
            )

        # Zero bound width
        with self.assertRaises(ValueError):
           res = (
               gr.Model()
               >> gr.cp_vec_function(
                   fun=lambda df: gr.df_make(
                       f=df.x**2 + df.y**2,
                       g=df.x + df.y,
                   ),
                   var=["x", "y"],
                   out=["f", "g"],
               )
               >> gr.cp_bounds(
                   x=( 0,  0),
                   y=(-1, +1),
               )

               >> gr.ev_contour(
                   var=["x", "y"],
                   out=["f", "g"],
                   n_side=10, # Coarse, for speed
               )
           )

        # No out
        with self.assertRaises(ValueError):
            res = (
                md1
                >> gr.ev_contour(var=["x", "y"])
            )

        # Output unavailable
        with self.assertRaises(ValueError):
            res = (
                md1
                >> gr.ev_contour(var=["x", "y"], out=["foo"])
            )



all = ["make_pareto_random"]

from grama import Model, cp_vec_function, cp_marginals, cp_copula_independence, \
    df_make, cos, sin, tan
from numpy import pi

def make_pareto_random(twoDim = True):
    """ Create a model of random points for a pareto frontier evaluation
    Args:
        twoDim (bool): determines whether to create a 2D or 3D model
    """
    if twoDim == True:
        # Model to make dataset
        md_true = (
            Model()
            >> cp_vec_function(
                fun=lambda df: df_make(
                    y1=df.x1 * cos(df.x2),
                    y2=df.x1 * sin(df.x2),
                ),
                var=["x1", "x2"],
                out=["y1", "y2"],
            )
            >> cp_marginals(
                x1=dict(dist="uniform", loc=0, scale=1),
                x2=dict(dist="uniform", loc=0, scale=pi/2),
            )
            >> cp_copula_independence()
        )

        return md_true
    else:
        # Model to make dataset
        md_true = (
            Model()
            >> cp_vec_function(
                fun=lambda df: df_make(
                    y1=df.x1 * cos(df.x2),
                    y2=df.x1 * sin(df.x2),
                    y3=df.x1 * tan(df.x2),
                ),
                var=["x1", "x2","x3"],
                out=["y1", "y2","y3"],
            )
            >> cp_marginals(
                x1=dict(dist="uniform", loc=0, scale=1),
                x2=dict(dist="uniform", loc=0, scale=pi/2),
                x3=dict(dist="uniform", loc=0, scale=pi/4)
            )
            >> cp_copula_independence()
        )

        return md_true



import unittest

from context import grama as gr
from context import data
from context import models

## Test autoplots
##################################################
class TestAutoplot(unittest.TestCase):

    def setUp(self):
        self.md = models.make_test()
        self.md_rand = (
            gr.Model()
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(
                    z=df.x + df.y,
                ),
                var=["x", "y"],
                out=["z"],
            )
            >> gr.cp_marginals(
                x=gr.marg_mom("norm", mean=0, sd=1),
                y=gr.marg_mom("norm", mean=0, sd=1),
            )
            >> gr.cp_copula_independence()
        )

        self.df_mc = gr.eval_sample(self.md, n=10, df_det="nom")
        self.df_mc_skip = gr.eval_sample(self.md, n=10, df_det="nom", skip=True)
        self.df_sinew = gr.eval_sinews(self.md, n_density=2, n_sweeps=1, df_det="nom")
        self.df_sinew_skip = gr.eval_sinews(
            self.md,
            n_density=2,
            n_sweeps=1,
            df_det="nom",
            skip=True
        )

        self.df_contour = gr.eval_contour(
            self.md,
            var=["x0", "x1"],
            out=["y0"],
            df=gr.df_make(x2=0),
        )
        self.df_contour_aux = gr.eval_contour(
            self.md,
            var=["x0", "x1"],
            out=["y0"],
            df=gr.df_make(x2=[0, 1]),
        )

        self.df_sobol = (
            self.md_rand
            >> gr.ev_hybrid(df_det="nom")
            >> gr.tf_sobol()
        )

    def test_autoplot(self):
        # Full color
        gr.plot_auto(self.df_mc)
        gr.plot_auto(self.df_mc_skip)
        gr.plot_auto(self.df_sinew)
        gr.plot_auto(self.df_sinew_skip)
        gr.plot_auto(self.df_contour)
        gr.plot_auto(self.df_sobol)
        # Black & White
        gr.plot_auto(self.df_mc, color="bw")
        gr.plot_auto(self.df_mc_skip, color="bw")
        gr.plot_auto(self.df_sinew, color="bw")
        gr.plot_auto(self.df_sinew_skip, color="bw")
        gr.plot_auto(self.df_contour, color="bw")
        gr.plot_auto(self.df_sobol, color="bw")

        ## iocorr fits into pipeline
        gr.plot_auto(
            self.md_rand
            >> gr.ev_sample(n=10, df_det="nom")
            >> gr.tf_iocorr()
        )

        with self.assertRaises(ValueError):
            gr.plot_auto(self.df_contour_aux)



import unittest
import io
import sys

from context import grama as gr
from numpy import eye, zeros
from numpy.random import multivariate_normal
from pandas import DataFrame

## Test support points
##################################################
class TestSupportPoints(unittest.TestCase):
    def setUp(self):
        n = 100
        self.df = DataFrame(
            data=multivariate_normal(zeros(3), eye(3), size=n), columns=["x", "y", "z"],
        )
        self.df["c"] = list(map(str, range(n)))

    def test_tran_sp(self):
        """Test the functionality of tran_sp()

        Note that *correctness* is verified elsewhere;
        see ./longrun/sp_convergence.ipynb
        """

        ## Basic facts
        df_sp = gr.tran_sp(self.df, n=10,)
        # Correct number of samples
        self.assertTrue(df_sp.shape[0] == 10)
        # Correct variables (numeric only)
        self.assertTrue(set(df_sp.columns) == {"x", "y", "z"})

        ## Subsetting target variables
        df_xy = gr.tran_sp(self.df, n=10, var=["x", "y"])
        # Subset correctly
        self.assertTrue(set(df_xy.columns) == {"x", "y"})

        ## Warning raised on lack of convergence
        with self.assertWarns(RuntimeWarning):
            gr.tran_sp(self.df, n=10, var=["x", "y"], n_maxiter=1)



#provides a class for dumping tab separated files for PGF
from __future__ import print_function
import matplotlib as mpl
from matplotlib.path import Path
from copy import deepcopy

class PGF:
	def __init__(self):
		self.column_names = []
		self.columns = []

	def add(self, name, column):
		if len(self.columns) > 1:
			assert len(self.columns[0]) == len(column)

		self.columns.append(deepcopy(column))
		self.column_names.append(name)

	def keys(self):
		return self.column_names

	def __getitem__(self, key):
		i = self.column_names.index(key)
		return self.columns[i]

	def write(self, filename):
		f = open(filename,'w')

		for name in self.column_names:
			f.write(name + '\t')
		f.write("\n")		

		for j in range(len(self.columns[0])):
			for col in self.columns:
				f.write("{}\t".format(float(col[j])))
			f.write("\n")

		f.close()

	def read(self, filename):
		with open(filename,'r') as f:
			for i, line in enumerate(f):
				# Remove the newline and trailing tab if present
				line = line.replace('\t\n','').replace('\n','')
				if i == 0:
					self.column_names = line.split('\t')
					self.columns = [ [] for name in self.column_names]
				else:
					cols = line.split('\t')
					for j, col in enumerate(cols):
						self.columns[j].append(float(col))


def save_contour(fname, cs, fmt = 'matlab', simplify = 1e-3, **kwargs):
	""" Save a contour plot to a file for pgfplots

	Additional arguments are passed to iter_segements
	Important, simplify = True will remove invisible points
	"""

	def write_path_matlab(fout, x_vec, y_vec, z):
		# Now dump this data back out
		# Header is level followed by number of rows
		fout.write('%15.15e\t%15d\n' % (z, len(x_vec)))
		for x, y in zip(x_vec, y_vec):
			fout.write("%15.15e\t%15.15e\n" % (x,y))

	def write_path_prepared(fout, x_vec, y_vec, z):
		fout.write("%15.15e\t%15.15e\t%15.15e\n" % (x0,y0,z))
		fout.write("\t\t\t\n")

	if fmt == 'matlab':
		write_path = write_path_matlab
	elif fmt == 'prepared':
		write_path = write_path_prepared
	else:
		raise NotImplementedError

	with open(fname, 'w') as fout:
		for col, z in zip(cs.collections, cs.levels):
			for path in col.get_paths():
				path.simplify_threshold = simplify
				x_vec = []
				y_vec = []
				for i, ((x,y), code) in enumerate(path.iter_segments(simplify = True)):
					if code == Path.MOVETO:
						if len(x_vec) !=0:
							write_path(fout, x_vec, y_vec, z)
							x_vec = []
							y_vec = []
						x_vec.append(x)
						y_vec.append(y)
					
					elif code == Path.LINETO:
						x_vec.append(x)
						y_vec.append(y)

					elif code == Path.CLOSEPOLY:
						x_vec.append(x_vec[0])
						y_vec.append(y_vec[0])
					else:
						print("received code", code)

				write_path(fout, x_vec, y_vec, z)



import numpy as np
import pandas as pd
import unittest
import io
import sys

from context import grama as gr
from context import models

## Test the Model Building Interface
##################################################
class TestMBI(unittest.TestCase):
    def setUp(self):
        self.md = gr.Model()

    def test_blank_model(self):
        """Checks that blank model is valid"""

        # Capture printpretty()
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        self.md.printpretty()
        sys.stdout = sys.__stdout__

        self.assertTrue(isinstance(self.md.domain, gr.Domain))

        self.assertTrue(isinstance(self.md.density, gr.Density))

    def test_getvars(self):
        # Define test function
        def fun(x, y, z):
            return x + y + z
        # Check that getvars function works correctly
        self.assertTrue(gr.getvars(fun), ("x", "y", "z"))

    def test_freeze(self):
        md_base = (
            gr.Model()
            >> gr.cp_vec_function(
                fun = lambda df: gr.df_make(
                    f=df.x + df.y + df.z
                ),
                var=["x", "y", "z"],
                out=["f"],
            )
            >> gr.cp_bounds(z=(-1, +1))
            >> gr.cp_marginals(
                x=gr.marg_mom("norm", mean=0, sd=1),
                y=gr.marg_mom("norm", mean=0, sd=1),
            )
            >> gr.cp_copula_gaussian(
                df_corr=gr.df_make(var1="x", var2="y", corr=0.5)
            )
        )

        md_frz = (
            md_base
            >> gr.cp_freeze(x=0)
        )

        # Correct variable list
        self.assertTrue(set(md_frz.var) == {"y", "z"})

        # Correct evaluation
        df_res = gr.eval_df(md_frz, gr.df_make(y=[0, 1], z=0))
        self.assertTrue(all(df_res["x"] == 0))

        # Can still sample
        df_samp = gr.eval_sample(md_frz, df_det="nom", n=10)

        # Freezing non-existent variables raises error
        with self.assertRaises(ValueError):
            gr.comp_freeze(md_base, foo=0)

        # Freezing with multiple-values raises error
        with self.assertRaises(ValueError):
            gr.comp_freeze(md_base, x=[0, 1])

        # Dataframe input
        md_frz_df = (
            md_base
            >> gr.cp_freeze(df=gr.df_make(x=0, z=0))
        )
        self.assertTrue(set(md_frz_df.var) == {"y"})
        df_res = gr.eval_sample(md_frz_df, n=10, df_det="nom")

    def test_comp_function(self):
        """Test comp_function()"""

        md_new0 = gr.comp_function(self.md, fun=lambda x: x, var=1, out=1)
        md_new1 = gr.comp_function(md_new0, fun=lambda x: x, var=1, out=1)

        ## Operations above should not affect self.md
        md_named = gr.comp_function(
            self.md,
            fun=lambda x: [x, 2 * x],
            var=["foo"],
            out=["bar1", "bar2"],
            name="test",
        )

        ## Default var and out names
        self.assertEqual(md_new0.var, ["x0"])
        self.assertEqual(md_new0.out, ["y0"])

        ## New default names iterate counter
        self.assertEqual(set(md_new1.var), set(["x0", "x1"]))
        self.assertEqual(set(md_new1.out), set(["y0", "y1"]))

        ## Output names assigned correctly
        # Also tests for copy issues
        self.assertEqual(set(md_named.out), set(["bar1", "bar2"]))
        ## Function name assigned correctly
        self.assertEqual(md_named.functions[0].name, "test")

        ## Invariant checks
        with self.assertRaises(ValueError):
            # Missing function
            gr.comp_function(self.md, fun=None, var=["foo"], out=["bar"])
        with self.assertRaises(ValueError):
            # Missing var
            gr.comp_function(self.md, fun=lambda x: x, var=None, out=["bar"])
        with self.assertRaises(ValueError):
            # Missing out
            gr.comp_function(self.md, fun=lambda x: x, var=["foo"], out=None)
        with self.assertRaises(ValueError):
            # Intersection var / out names
            self.md >> gr.cp_function(lambda x: x, var=["x"], out=["x"], name="f0")
        with self.assertRaises(ValueError):
            # Non-unique function names
            self.md >> gr.cp_function(
                lambda x: x, var=1, out=1, name="f0"
            ) >> gr.cp_function(lambda x: x, var=1, out=1, name="f0")
        ## DAG invariant checks
        with self.assertRaises(ValueError):
            # Cycle by input
            self.md >> gr.cp_function(
                fun=lambda x0: x0, var=["y0"], out=1
            ) >> gr.cp_function(fun=lambda x0: x0, var=1, out=["y0"])
        with self.assertRaises(ValueError):
            # Non-unique output
            self.md >> gr.cp_function(
                fun=lambda x0: x0, var=1, out=["y0"]
            ) >> gr.cp_function(fun=lambda x0: x0, var=1, out=["y0"])

        ## Check vectorized builder
        md_vec = gr.comp_vec_function(
            self.md, fun=lambda df: df.assign(y0=df.x0), var=1, out=1,
        )
        self.assertTrue(
            gr.df_equal(gr.df_make(x0=0, y0=0), md_vec >> gr.ev_df(df=gr.df_make(x0=0)))
        )

    def test_comp_model(self):
        """Test model composition"""
        md_inner = (
            gr.Model("inner")
            >> gr.cp_function(fun=lambda x0, x1: x0 + x1, var=2, out=1)
            >> gr.cp_marginals(x0=dict(dist="norm", loc=0, scale=1))
            >> gr.cp_copula_independence()
        )

        ## Deterministic composition
        md_det = gr.Model("outer_det") >> gr.cp_md_det(md=md_inner)

        self.assertTrue(set(md_det.var) == {"x0", "x1"})
        self.assertTrue(md_det.out == ["y0"])
        gr.eval_df(md_det, df=gr.df_make(x0=0, x1=0))

        ## Deterministic composition
        md_sample = gr.Model("outer_det") >> gr.cp_md_sample(
            md=md_inner, param=dict(x0=("loc", "scale"))
        )

        self.assertTrue(set(md_sample.var) == {"x0_loc", "x0_scale", "x1"})
        self.assertTrue(set(md_sample.out) == {"y0"})
        gr.eval_df(md_sample, df=gr.df_make(x0_loc=0, x0_scale=1, x1=0))

    def test_comp_bounds(self):
        """Test comp_bounds()"""

        ## Add bound w/o function
        md1 = gr.comp_bounds(self.md, x=(0, 1))

        self.assertEqual(md1.domain.bounds["x"], [0, 1])

        ## Add bound with function
        md2 = gr.comp_function(self.md, fun=lambda x: x, var=["x"], out=["y"])
        md2 = gr.comp_bounds(md2, x=(0, 1))

        self.assertEqual(md2.n_var, 1)

        ## Add multiple bounds
        md3 = gr.comp_function(self.md, fun=lambda x: x, var=2, out=2)
        md3 = gr.comp_bounds(md3, x0=(0, 1), x1=(0, 1))

        self.assertEqual(md3.n_var, 2)

        ## Additional bound values ignored
        md4 = gr.comp_bounds(self.md, x=(0, 1, 2))

        self.assertEqual(md4.domain.bounds["x"][0], 0)
        self.assertEqual(md4.domain.bounds["x"][1], 1)

    def test_comp_marginals(self):
        """Test comp_bounds()"""

        ## Add marginal w/o function
        md1 = gr.comp_marginals(self.md, x={"dist": "norm", "loc": 0, "scale": 1})

        self.assertEqual(md1.density.marginals["x"].d_name, "norm")

        ## Check sign and default sign
        md2 = gr.comp_marginals(
            self.md,
            x={"dist": "norm", "loc": 0, "scale": 1},
            y={"dist": "norm", "loc": 0, "scale": 1, "sign": -1},
        )

        self.assertEqual(md2.density.marginals["x"].sign, 0)
        self.assertEqual(md2.density.marginals["y"].sign, -1)

        ## Check number and list of vars computed correctly
        md3 = gr.comp_bounds(self.md, x=(0, 1), y=(0, 1))
        md3 = gr.comp_marginals(md3, x={"dist": "uniform", "loc": 0, "scale": 1})

        self.assertEqual(md3.n_var, 2)
        self.assertEqual(md3.n_var_det, 1)
        self.assertEqual(md3.n_var_rand, 1)
        self.assertEqual(md3.var_det, ["y"])
        self.assertEqual(md3.var_rand, ["x"])

        ## Invariant raises
        with self.assertRaises(NotImplementedError):
            gr.comp_marginals(self.md, x={})

    def test_comp_copula(self):
        md_incomplete = gr.comp_marginals(
            self.md,
            x={"dist": "uniform", "loc": -1, "scale": 2},
            y={"dist": "uniform", "loc": -1, "scale": 2},
        )

        with self.assertRaises(AttributeError):
            md_incomplete.sample()

        ## Independence copula
        md_independence = gr.comp_copula_independence(md_incomplete)
        df_ind = md_independence.density.sample()
        self.assertTrue(set(df_ind.columns) == set(["x", "y"]))

        ## Gaussian copula
        md_gaussian = gr.comp_copula_gaussian(
            md_incomplete, pd.DataFrame(dict(var1=["x"], var2=["y"], corr=[0.5]))
        )
        df_gau = md_gaussian.density.sample()
        self.assertTrue(set(df_gau.columns) == set(["x", "y"]))



import unittest
import io
import sys

from context import grama as gr

DF = gr.Intention()

## Test support points
##################################################
class TestPolyridge(unittest.TestCase):
    def setUp(self):
        pass

    def test_fit_polyridge(self):
        """Test the functionality and correctness of ft_polyridge()
        """
        df_test = (
            gr.df_make(x=range(10))
            >> gr.tf_outer(gr.df_make(y=range(10)))
            >> gr.tf_outer(gr.df_make(z=range(10)))
            >> gr.tf_mutate(f=DF.x - DF.y)
        )

        md = gr.fit_polyridge(df_test, out="f", n_degree=1, n_dim=1)

        df1 = gr.eval_df(md, df=gr.df_make(x=[2,1], y=[1], z=[0]))
        df2 = gr.df_make(x=[2,1], y=[1], z=[0], f_mean=[1, 0])

        self.assertTrue(gr.df_equal(
            df1,
            df2,
            close=True,
        ))

    def test_tran_polyridge(self):
        """Test the functionality and correctness of tran_polyridge()
        """
        ## Setup
        df_test = (
            gr.df_make(x=range(10))
            >> gr.tf_outer(gr.df_make(y=range(10)))
            >> gr.tf_outer(gr.df_make(z=range(10)))
            >> gr.tf_mutate(f=DF.x - DF.y)
        )

        ## Assertions
        # No `out` column
        with self.assertRaises(ValueError):
            gr.tran_polyridge(df_test)
        # Unrecognized `out` column
        with self.assertRaises(ValueError):
            gr.tran_polyridge(df_test, out="foo")
        # Unrecognized `var` column(s)
        with self.assertRaises(ValueError):
            gr.tran_polyridge(df_test, var=["foo", "bar"])
        # Invalid degree
        with self.assertRaises(ValueError):
            gr.tran_polyridge(df_test, out="f", n_degree=1, n_dim=2)

        ## Correctness
        df_res = (
            df_test
            >> gr.tf_polyridge(
                out="f",
                n_dim=1,
                n_degree=1,
            )
        )
        df_true = gr.df_make(x=1/gr.sqrt(2), y=-1/gr.sqrt(2), z=0)

        self.assertTrue(gr.df_equal(df_res, df_true, close=True))

        ## Higher-dimensional case runs without error
        df_higher = (
            gr.df_grid(
                x=range(10),
                y=range(10),
                z=range(10),
            )
            >> gr.tf_mutate(f=DF.x + DF.y + DF.z)
        )
        gr.tran_polyridge(df_higher, out="f", n_degree=2, n_dim=2)

        ## Fitting seed runs without error
        gr.fit_polyridge(df_higher, out="f", n_degree=1, n_dim=1, seed=101)

        ## Seed stabilizes results
        df_res1 = gr.tran_polyridge(df_higher, out="f", n_degree=1, n_dim=1, seed=101)
        df_res2 = gr.tran_polyridge(df_higher, out="f", n_degree=1, n_dim=1, seed=101)

        self.assertTrue(gr.df_equal(df_res1, df_res2, close=True))

## Run tests
if __name__ == '__main__':
    import xmlrunner

    unittest.main(
        testRunner=xmlrunner.XMLTestRunner(output='test-reports'),
        failfast=False,
        buffer=False,
        catchbreak=False,
    )



r""" Tools for computing Voronoi vertices
"""

from __future__ import print_function

import numpy as np
from scipy.spatial.distance import cdist, pdist
import scipy.linalg
from cvxopt import matrix, solvers
solvers.options['show_progress'] = False

from scipy.spatial import HalfspaceIntersection
from .cdist import cdist
from .geometry import unique_points

from scipy.spatial.qhull import QhullError



def voronoi_vertex_1d(domain, Xhat, L = None):
	r""" One dimensional voronoi vertices

	As Qhull only works on two or more dimensions, 
	this code handles the simple, one-dimensional case
	"""
	assert len(domain) == 1, "Domain must be one-dimensional"
	
	lb = domain.corner(np.array([-1]))
	ub = domain.corner(np.array([1]))
	
	X = np.sort(np.hstack([Xhat.flatten(), ub, lb]))
	# mid point between X
	V = 0.5*(X[1:] + X[:-1])
	# add boundaries
	V = np.hstack([lb, V, ub])
	return V.reshape(-1,1)

def voronoi_vertex(domain, Xhat, L = None):
	r""" Construct the bounded Voronoi vertices on a domain 




	Note: 
	Pro17 claims that these bounded Voronoi vertices can be computed using WP89.  
	This approach may offer some speedup, but I trust Q-hull more than my own implementation.

	"""
	if len(domain) == 1:
		return voronoi_vertex_1d(domain, Xhat)
	
	assert np.all(domain.isinside(Xhat))


	if L is None:
		L = np.eye(len(domain))
		Linv = np.eye(len(domain))
	else:
		U, s, VT = np.linalg.svd(L)
		Linv = VT.T @ np.diag(1./s) @ U.T



	# Half spaces from the domain
	A, b = domain.A_aug, domain.b_aug

	if len(Xhat) == 1:
		halfspaces = np.hstack([A, -b.reshape(-1,1)])
		hs = HalfspaceIntersection(halfspaces, Xhat[0])
		return hs.intersections	

	Yhat = (L @ Xhat.T).T
	LV = []
	for k, xhat in enumerate(Xhat):
		Ak = Yhat - Yhat[k] 
		#Ak = np.vstack([L @ (Xhat[j] - xhat) for j in range(len(Xhat)) if j != k])
		center = 0.5 * (Yhat[k] + Yhat)
		#center = np.vstack([ 0.5*L @ (xhat + Xhat[j]) for j in range(len(Xhat)) if j != k]) 
		bk = np.sum(Ak * center, axis = 1)
		
		# If two Xhat are the same, we can end up with a situation 
		# where the constraint stops  so we filter these out
		I = np.argwhere(np.sum(Ak**2, axis = 1) > 0).flatten()
		Ak, bk = Ak[I], bk[I]
	
		halfspaces = np.hstack([np.vstack([A @ Linv , Ak  ]) , -np.hstack([b, bk]).reshape(-1,1)])

		try:
			hs = HalfspaceIntersection(halfspaces, xhat)
		except QhullError:
			# The point xhat isn't strictly inside the constraints, so we try with a point that is
			# Here we use the code from scipy documentation 
			norm_vector = np.reshape(np.linalg.norm(halfspaces[:, :-1], axis=1),(halfspaces.shape[0], 1))
			cc = np.zeros((halfspaces.shape[1],))
			cc[-1] = -1
			AA = np.hstack((halfspaces[:, :-1], norm_vector))
			bb = -halfspaces[:, -1:]
			
			# Solve the linear program with CVXOPT as scipy's linprog reports ill-conditioning 
			sol = solvers.lp(matrix(cc), matrix(AA), matrix(bb))
			xhat2 = np.array(sol['x'][:-1]).flatten()
			
			try:
				hs = HalfspaceIntersection(halfspaces, xhat2)
			except QhullError as e:
				raise e


		# TODO: Do we want to keep indexing information, or only return the vertices 
		LV.append(np.copy(hs.intersections))

	LV = np.vstack(LV)
	V = (Linv @ LV.T).T
	I = unique_points(V)
	V = V[I]
	
	I = domain.isinside(V)
	V = V[I]
	return V

def voronoi_vertex_sample(domain, Xhat, X0, L = None, randomize = True):
	r""" Constructs a subset of the Voronoi vertices on a given domain 


	Given a domain :math:`\mathcal{D} \subset \mathbb{R}^m`, a set of points
	:math:`\lbrace \widehat{\mathbf{x}}_j \rbrace_{j=1}^M`, and a distance metric
	
	.. math::
		
		d(\mathbf{x}_1, \mathbf{x}_2) = \|\mathbf{L}(\mathbf{x}_1 - \mathbf{x}_2)\|_2

	the vertices of the bounded Voronoi diagram :math:`\mathbf{v}_i \in \mathcal{D}`
	are those points are those points that satisfy :math:`r` equidistance constraints
	in the metric :math:`d`

	.. math::

		\| \mathbf{L}(\mathbf{v}_i - \widehat{\mathbf{x}}_{\mathcal{I}_i[1]})\|_2 =
		\| \mathbf{L}(\mathbf{v}_i - \widehat{\mathbf{x}}_{\mathcal{I}_i[2]})\|_2 = \ldots = 
		\| \mathbf{L}(\mathbf{v}_i - \widehat{\mathbf{x}}_{\mathcal{I}_i[r]})\|_2 

	and :math:`m-r` constraints from the domain.
	
	
	The bounded Voronoi vertices include the vertices of the domain.
	The number of domain vertices can grow exponentially in dimension (e.g., consider the cube),
	but the asymptotic cost of finding these is :math:`\mathcal{O}(mn)` [wikiVEP]_ 
	per vertex where :math:`n` is the number of linear inequality constraints specifying the domain.
	Hence, the complexity of finding the Voronoi vertices grows exponentially in the 
	dimension of the domain.
	Thus we use an approach due Lindemann and Cheng [LC05]_ to find a subset of vertices
	by taking a starting point :math:`\mathbf{x}_0` and sequentially projecting it
	onto a series of hyperplanes such that after :math:`m` steps we satisfy :math:`m` constraints.	

	Parameters
	----------
	domain: Domain
		Domain on which to construct the vertices
	Xhat: array-like (M, m)
		M existing points on the domain which define the Voronoi diagram
	X0: array-like (N, m)
		Initial points to use to find vertices
	L: array-like (m, m), optional
		Weight on distance in 2-norm; defaults to the identity matrix	
	randomize: bool, optional (default: True)
		If true, when L is rank deficient add a random direction at each step
		such that we find points on the boundary of the domain satisfying m
		constraints.


	Returns
	-------
	X: np.ndarray(N, m)
		Points satisfying m constraints


	References
	----------
	.. [LC05] Iteratively Locating Voronoi Vertices for Dispersion Estimation
		Stephen R. Lindemann and Peng Cheng
		Proceedings of the 2005 Interational Conference on Robotics and Automation
	.. [wikiVEP] Vertex enumeration problem, Wikipedia.
		https://en.wikipedia.org/wiki/Vertex_enumeration_problem
	"""

	# Startup checks
	#assert isinstance(domain, Domain), "domain must be an instance of the Domain class"
	assert len(domain.Ls) == 0, "Currently this does not support domains with quadratic constraints"
	Xhat = np.atleast_2d(np.array(Xhat))
	X0 = np.atleast_2d(np.array(X0)).copy()

	m = len(domain)
	
	if L is None:
		L = np.eye(m)
	else:
		L = np.array(L)
		assert L.shape[1] == m, "Number of columns doesn't match dimension of the space"

	Lrank = np.linalg.matrix_rank(L)

	LTL = L.T.dot(L)

	# Linear inequality constraints for the domain (including lower bound/upper bound box constraints)
	A = domain.A_aug
	b = domain.b_aug

	# This algorithm terminates when each point has m active constraints as we can have no more improvement
	# hence we substract the number of equality constraints
	# we also subtract the rank-deficency of L because this results in zero-steps  	
	for k in range(m - domain.A_eq.shape[0]):
		if k >= Lrank and not randomize:
			break

		# Find the nearest neighbors
		# As we intend to do this in high dimensions, 
		# we don't use a tree-based distance approach
		# as these don't scale well
		D = cdist(L.dot(X0.T).T, L.dot(Xhat.T).T)
		I = np.argsort(D, axis = 1)

		# set the search direction to move away from the closest point
		h = X0 - Xhat[I[:,0]]
		h = LTL.dot(h.T).T
			 

		# which inequality constraints on the domain are active
		active = np.isclose(A.dot(X0.T).T , b)
		
		for i in range(X0.shape[0]):
			# For each point, project onto the feasible directions
			nullspace = [domain.A_eq.T]

			# active inequality constraints
			for j in range(A.shape[0]):
				if active[i,j] : nullspace += [A[j].reshape(-1,1)]

			# constraints from other points in the domain
			for j in range(1, min(k+1, len(Xhat))):
				if np.isclose(D[i,I[i,0]], D[i,I[i,j]]):	
					nullspace += [LTL.dot(Xhat[I[i,0]] - Xhat[I[i,j]]).reshape(-1,1)]
					#h[i] += (X0[i] - Xhat[I[i,j]])

			nullspace = np.hstack(nullspace)
			# If there are no active constraints, don't do anything
			if nullspace.shape[1]>0:
				Q, R = scipy.linalg.qr(nullspace, overwrite_a = True, mode = 'economic')
				h[i] -= Q.dot(Q.T.dot(h[i]))
				if np.isclose(np.linalg.norm(h[i]), 0) and randomize:
					# If L is low-rank we can have situations where h[i]	
					# as constructed above is in the nullspace of constraints
					# and hence is approximately zero.  When randomize=True
					# we choose a random search direction so we can still make 
					# progress towards satisfying m constraints.
					h[i] = np.random.randn(*h[i].shape)
					h[i] -= Q.dot(Q.T.dot(h[i]))
				#print("k=%d, constraints %d, norm %g" % (k, Q.shape[1], np.linalg.norm(h[i])))
		
		# Now we find the furthest we can step along this direction before either hitting a
		# (1) separating hyperplane separating x0 and points in Xhat or 
		# (2) the boundary of the domain

		alpha = np.inf*np.ones(X0.shape[0])

		# (1) Take a step until we run into a separating hyperplane
		for xhat in Xhat:
			# we setup a hyperplane (p - p0)^* n  = 0 
			# where p0 is a point on the hyperplane
			# separating the closest point Xhat[:,I[:,0]] and xhat
			p0 = 0.5*(L.dot( (xhat + Xhat[I[:,0]]).T).T)
			n = L.dot( (xhat - Xhat[I[:,0]]).T).T

			# Inner product of normal n with search direction h
			nh = np.sum(n*L.dot(h.T).T, axis = 1)

			# Inner product for the numerator (x0 - p0)^* n
			numerator = -np.sum( (L.dot(X0.T).T - p0)*n, axis = 1)
			
			alpha_c = np.inf*np.ones(alpha.shape)
			act = ~np.isclose(np.abs(nh), 0) 
			alpha_c[act] = numerator[act]/nh[act]
			# When xhat is the closest point, we get a divide by zero error
			# We cannot move backwards, so these are also set to infinity
			alpha_c[alpha_c < 0 ] = np.inf
			alpha = np.minimum(alpha, alpha_c)

		# (2) Find intersection with the domain
		# cf., Domain._extent_ineq 
		AX0 = A.dot(X0.T).T
		Ah = A.dot(h.T).T
		with np.errstate(divide = 'ignore', invalid = 'ignore'):
			alpha_c = (b - AX0)/Ah
		alpha_c[~np.isfinite(alpha_c)] = np.inf
		alpha_c[alpha_c <= 0] = np.inf
		# When the step size is near zero, the above formula is inaccurate
		# causing points to emerge outside of the domain. Hence, we zero the
		# step size for these points.
		alpha_c[ np.max(np.abs(h), axis = 1) < 1e-10] = 0. 
		alpha_c = np.min(alpha_c, axis = 1)
		alpha = np.minimum(alpha, alpha_c)

		# There are some cases above where both alpha and alpha_c will yield an infinite step
		# in this case we zero these just to be safe
		alpha[~np.isfinite(alpha)] = 0.

		# Now finally take the step
		X0 += alpha.reshape(-1,1)*h

	
	# Keep only those points that are inside the domain
	# (all points should be, but sometimes numerical issues push outside)
	#X0 = X0[domain.isinside(X0)]

	return X0



"""Ridge function approximation from function values"""
# (c) 2017 Jeffrey M. Hokanson (jeffrey@hokanson.us)

import numpy as np

from .function import BaseFunction
from .subspace import SubspaceBasedDimensionReduction
from .pgf import PGF

class RidgeFunction(BaseFunction, SubspaceBasedDimensionReduction):
	@property
	def U(self):
		return self._U

	def shadow_plot(self, X = None, fX = None, dim = None, ax = 'auto', pgfname = None, **kwargs):
		if dim is None:
			dim = self.U.shape[1]
		else:
			assert dim == self.U.shape[1]

		ax = SubspaceBasedDimensionReduction.shadow_plot(self, X = X, fX = fX, dim = dim, ax = ax, pgfname = pgfname)

		# Draw the response surface
		if dim == 1:
			Y = np.dot(self.U.T, X.T).T
			lb = np.min(Y)
			ub = np.max(Y)
			
			xx = np.linspace(lb, ub, 500)
			Uxx = np.hstack([self.U*xxi for xxi in xx]).T
			yy = self.eval(Uxx)

			if ax is not None:
				ax.plot(xx, yy, 'r-')

			if pgfname is not None:	
				pgfname2 = pgfname[:pgfname.rfind('.')] + '_response' + pgfname[pgfname.rfind('.'):]
				pgf = PGF()
				pgf.add('x', xx)
				pgf.add('fx', yy )
				pgf.write(pgfname2)
					

		elif dim == 2:	
			Y = np.dot(self.U.T, X.T).T
			lb0 = np.min(Y[:,0])
			ub0 = np.max(Y[:,0])

			lb1 = np.min(Y[:,1])
			ub1 = np.max(Y[:,1])

			# Constuct mesh on the domain
			xx0 = np.linspace(lb0, ub0, 50)
			xx1 = np.linspace(lb1, ub1, 50)
			XX0, XX1 = np.meshgrid(xx0, xx1)
			UXX = np.vstack([XX0.flatten(), XX1.flatten()])
			XX = np.dot(self.U, UXX).T
			YY = self.eval(XX).reshape(XX0.shape)
			
			ax.contour(xx0, xx1, YY, 
				levels = np.linspace(np.min(fX), np.max(fX), 20), 
				vmin = np.min(fX), vmax = np.max(fX),
				linewidths = 0.5)
		

		else: 
			raise NotImplementedError("Cannot draw shadow plots in more than two dimensions")	

		return ax		





__all__ = [
    "tran_mutate",
    "tf_mutate",
    "tran_mutate_if",
    "tf_mutate_if",
]

from .base import dfdelegate, make_symbolic, flatten
from .. import add_pipe


@dfdelegate
def tran_mutate(df, **kwargs):
    """Create new columns

    Creates new variables (columns) in the DataFrame specified by keyword argument pairs, where the key is the column name and the value is the new column value(s).

    Use the Intention operator (usually `DF = gr.Intention()`) as a convenient way to access columns in the DataFrame.

    Args:
        df (pandas.DataFrame): Data to modify

    Kwargs:
        **kwargs: Compute new column values
            the name of the argument (left of `=`) will be the new column name,
            the value of the argument (right of `=`) defines the new column's value

    Examples::

        ## Setup
        import grama as gr
        DF = gr.Intention()
        ## Load example dataset
        from grama.data import df_diamonds

        ## Compute a rough estimate of volume
        (
            df_diamonds
            >> gr.tf_mutate(
                vol=DF.x * DF.y * DF.z
            )
            >> gr.tf_select("x", "y", "z", "vol")
        )

    """

    return df.assign(**kwargs)

tf_mutate = add_pipe(tran_mutate)


@dfdelegate
def tran_mutate_if(df, predicate, fun):
    """
    Modifies columns in place if the specified predicate is true.
    Args:
        df (pandas.DataFrame): data passed in through the pipe.
        predicate: a function applied to columns that returns a boolean value
        fun: a function that will be applied to columns where predicate returns True

    Examples::

        diamonds >> mutate_if(lambda col: min(col) < 1 and mean(col) < 4, lambda row: 2 * row) >> head(3)
           carat      cut color clarity  depth  table  price     x     y     z
        0   0.46    Ideal     E     SI2   61.5   55.0    326  3.95  3.98  4.86
        1   0.42  Premium     E     SI1   59.8   61.0    326  3.89  3.84  4.62
        2   0.46     Good     E     VS1   56.9   65.0    327  4.05  4.07  4.62
        (columns 'carat' and 'z', both having a min < 1 and mean < 4, are doubled, while the
        other rows remain as they were)
    """
    cols = list()
    for col in df:
        try:
            if predicate(df[col]):
                cols.append(col)
        except:
            pass
    df[cols] = df[cols].apply(fun)
    return df

    # df2 = df.copy()
    # df2[cols] = df2[cols].apply(fun)
    # return df2

tf_mutate_if = add_pipe(tran_mutate_if)



# Statistical Process Control tools
__all__ = [
    "c_sd",
    "B3",
    "B4",
    "plot_xbs",
    "pt_xbs",
]

from scipy.special import gamma
from numpy import sqrt
from grama import n as nfcn

from grama import (
    add_pipe,
    Intention,
    tf_group_by,
    tf_summarize,
    tf_mutate,
    tf_ungroup,
    tf_filter,
    tf_pivot_longer,
    tf_left_join,
    mean,
    sd,
    lead,
    lag,
    consec,
    case_when,
    theme_uqbook,
)

from plotnine import (
    ggplot,
    aes,
    geom_line,
    geom_hline,
    geom_point,
    element_rect,
    facet_grid,
    theme,
    theme_minimal,
    guides,
    scale_linetype_manual,
    scale_shape_manual,
    scale_color_manual,
    labs,
    labeller,
)

## Helper functions
# --------------------------------------------------
def c_sd(n):
    r"""Anti-biasing constant for aggregate standard deviation

    Returns the anti-biasing constant for aggregated standard deviation estimates. If the average of $k$ samples each size $n$ are averaged to produce $\overline{S} = (1/k) \sum_{i=1}^k S_i$, then the de-biased standard deviation is:

        $$\hat{\sigma} = \overline{S} / c(n)$$

    Arguments:
        n (int): Sample (batch) size

    Returns:
        float: anti-biasing constant

    References:
        Kenett and Zacks, Modern Industrial Statistics (2014) 2nd Ed

    """
    return gamma(n/2) / gamma( (n-1)/2 ) * sqrt( 2 / (n-1) )


def B3(n):
    r"""Lower Control Limit constant for standard deviation

    Returns the Lower Control Limit (LCL) constant for aggregated standard
    deviation estimates. If the average of $k$ samples each size $n$ are
    averaged to produce $\overline{S} = (1/k) \sum_{i=1}^k S_i$, then the LCL
    is:

        $$LCL = B_3 \overline{S}$$

    Arguments:
        n (int): Sample (batch) size

    Returns:
        float: LCL constant

    References:
        Kenett and Zacks, Modern Industrial Statistics (2014) 2nd Ed, Equation (8.22)

    """
    return max( 1 - 3 / c_sd(n) * sqrt(1 - c_sd(n)**2), 0 )


def B4(n):
    r"""Upper Control Limit constant for standard deviation

    Returns the Upper Control Limit (UCL) constant for aggregated standard
    deviation estimates. If the average of $k$ samples each size $n$ are
    averaged to produce $\overline{S} = (1/k) \sum_{i=1}^k S_i$, then the UCL
    is:

        $$UCL = B_4 \overline{S}$$

    Arguments:
        n (int): Sample (batch) size

    Returns:
        float: UCL constant

    References:
        Kenett and Zacks, Modern Industrial Statistics (2014) 2nd Ed, Equation (8.22)

    """
    return 1 + 3 / c_sd(n) * sqrt(1 - c_sd(n)**2)


## Control Chart constructors
# --------------------------------------------------
def plot_xbs(df, group, var, n_side=9, n_delta=6, color="full"):
    r"""Construct Xbar and S chart

    Construct an Xbar and S chart to assess the state of statistical control of
    a dataset.

    Args:
        df (DataFrame): Data to analyze
        group (str): Variable for grouping
        var (str): Variable to study

    Keyword args:
        n_side (int): Number of consecutive runs above/below centerline to flag
        n_delta (int): Number of consecutive runs increasing/decreasing to flag
        color (str): Color mode; "full" for full color, "bw" for greyscale

    Returns:
        plotnine object: Xbar and S chart

    Examples::

        import grama as gr
        DF = gr.Intention()

        from grama.data import df_shewhart
        (
            df_shewhart
            >> gr.tf_mutate(idx=DF.index // 10)
            >> gr.pt_xbs("idx", "tensile_strength")
        )

    """
    ## Prepare the data
    DF = Intention()
    df_batched = (
        df
        >> tf_group_by(group)
        >> tf_summarize(
            X=mean(DF[var]),
            S=sd(DF[var]),
            n=nfcn(DF.index),
        )
        >> tf_ungroup()
    )

    df_stats = (
        df_batched
        >> tf_summarize(
            X_center=mean(DF.X),
            S_biased=mean(DF.S),
            n=mean(DF.n),
        )
    )
    n = df_stats.n[0]
    df_stats["S_center"] = df_stats.S_biased / c_sd(n)
    df_stats["X_LCL"] = df_stats.X_center - 3 * df_stats.S_center / sqrt(n)
    df_stats["X_UCL"] = df_stats.X_center + 3 * df_stats.S_center / sqrt(n)
    df_stats["S_LCL"] = B3(n) * df_stats.S_center
    df_stats["S_UCL"] = B4(n) * df_stats.S_center

    ## Reshape for plotting
    df_stats_long = (
        df_stats
        >> tf_pivot_longer(
            columns=["X_LCL", "X_center", "X_UCL", "S_LCL", "S_center", "S_UCL"],
            names_to=["_var", "_stat"],
            names_sep="_",
            values_to="_value",
        )
    )
    # Fake group value to avoid issue with discrete group variable
    df_stats_long[group] = [df_batched[group].values[0]] * df_stats_long.shape[0]

    df_batched_long = (
        df_batched
        >> tf_pivot_longer(
            columns=["X", "S"],
            names_to="_var",
            values_to="_value",
        )
        ## Flag patterns
        >> tf_left_join(
            df_stats
            >> tf_pivot_longer(
                columns=["X_LCL", "X_center", "X_UCL", "S_LCL", "S_center", "S_UCL"],
                names_to=["_var", ".value"],
                names_sep="_",
            ),
            by="_var",
        )
        >> tf_group_by("_var")
        >> tf_mutate(
            outlier_below=(DF._value < DF.LCL),        # Outside control limits
            outlier_above=(DF.UCL < DF._value),
            below=consec(DF._value < DF.center, i=n_side), # Below mean
            above=consec(DF.center < DF._value, i=n_side), # Above mean
        )
        >> tf_mutate(
            decreasing=consec((lead(DF._value) - DF._value) < 0, i=n_delta-1) | # Decreasing
                       consec((DF._value - lag(DF._value)) < 0, i=n_delta-1),
            increasing=consec(0 < (lead(DF._value) - DF._value), i=n_delta-1) | # Increasing
                       consec(0 < (DF._value - lag(DF._value)), i=n_delta-1),
        )
        >> tf_mutate(
            sign=case_when(
                [DF.outlier_below, "-2"],
                [DF.outlier_above, "+2"],
                [DF.below | DF.decreasing, "-1"],
                [DF.above | DF.increasing, "+1"],
                [True, "0"]
            ),
            glyph=case_when(
                [DF.outlier_below, "Below Limit"],
                [DF.outlier_above, "Above Limit"],
                [DF.below, "Low Run"],
                [DF.above, "High Run"],
                [DF.increasing, "Increasing Run"],
                [DF.decreasing, "Decreasing Run"],
                [True, "None"],
            )
        )
        >> tf_ungroup()
    )

    ## Visualize
    if color == "full":
        return (
            df_batched_long
            >> ggplot(aes(x=group))
            + geom_hline(
                data=df_stats_long,
                mapping=aes(yintercept="_value", linetype="_stat"),
            )
            + geom_line(aes(y="_value", group="_var"), size=0.2)
            + geom_point(
                aes(y="_value", color="sign", shape="glyph"),
                size=3,
            )

            + scale_color_manual(
                values={"-2": "blue", "-1": "darkturquoise", "0": "black", "+1": "salmon", "+2": "red"},
            )
            + scale_shape_manual(
                name="Patterns",
                values={
                    "Below Limit": "X",
                    "Above Limit": "X",
                    "Low Run": "s",
                    "High Run": "s",
                    "Increasing Run": "^",
                    "Decreasing Run": "v",
                    "None": "."
                },
            )
            + scale_linetype_manual(
                name="Guideline",
                values=dict(LCL="dashed", UCL="dashed", center="solid"),
            )
            + theme_uqbook()
            + theme(panel_border=element_rect(color="black", fill=None, size=0.5))
            + guides(color=None)
            + facet_grid(
                "_var~.",
                scales="free_y",
                labeller=labeller(dict(X="Mean", S="Variability")),
            )
            + labs(
                x="Group variable ({})".format(group),
                y="Value ({})".format(var),
            )
        )
    elif color == "bw":
        return (
            df_batched_long
            >> ggplot(aes(x=group))
            + geom_hline(
                data=df_stats_long,
                mapping=aes(yintercept="_value", linetype="_stat"),
            )
            + geom_line(aes(y="_value", group="_var"), size=0.2)
            + geom_point(
                aes(y="_value", color="sign", shape="glyph"),
                size=3,
            )

            + scale_color_manual(
                values={
                    "-2": "grey",
                    "-1": "grey",
                    "0": "black",
                    "+1": "grey",
                    "+2": "grey"
                },
            )
            + scale_shape_manual(
                name="Patterns",
                values={
                    "Below Limit": "X",
                    "Above Limit": "X",
                    "Low Run": "s",
                    "High Run": "s",
                    "Increasing Run": "^",
                    "Decreasing Run": "v",
                    "None": "."
                },
            )
            + scale_linetype_manual(
                name="Guideline",
                values=dict(LCL="dashed", UCL="dashed", center="solid"),
            )
            + theme_uqbook()
            + theme(panel_border=element_rect(color="black", fill=None, size=0.5))
            + guides(color=None)
            + facet_grid(
                "_var~.",
                scales="free_y",
                labeller=labeller(dict(X="Mean", S="Variability")),
            )
            + labs(
                x="Group variable ({})".format(group),
                y="Value ({})".format(var),
            )
        )


pt_xbs = add_pipe(plot_xbs)



r""" LinQuadDomain definition

Philosophically, LinQuad domains consist of those convex domains specified by a combination
of linear inequality, linear equality, and quadratic inequality constraints.

From a code perspective, this is where a dependency on CVXPY is introduced to
handle domain properities from within this domain

"""

from __future__ import division

import numpy as np
from .domain import TOL#, DEFAULT_CVXPY_KWARGS
from .euclidean import EuclideanDomain
from .tensor import TensorProductDomain

from ..misc import merge

class LinQuadDomain(EuclideanDomain):
	r"""A domain specified by a combination of linear (in)equality constraints and convex quadratic constraints


	Here we define a domain that is specified in terms of bound constraints,
	linear inequality constraints, linear equality constraints, and quadratic constraints.

	.. math::

		\mathcal{D} := \left \lbrace
			\mathbf{x} : \text{lb} \le \mathbf{x} \le \text{ub}, \
			\mathbf{A} \mathbf{x} \le \mathbf{b}, \
			\mathbf{A}_{\text{eq}} \mathbf{x} = \mathbf{b}_{\text{eq}}, \
			\| \mathbf{L}_i (\mathbf{x} - \mathbf{y}_i)\|_2 \le \rho_i
		\right\rbrace \subset \mathbb{R}^m


	Parameters
	----------
	A: array-like (m,n)
		Matrix in left-hand side of inequality constraint
	b: array-like (m,)
		Vector in right-hand side of the ineqaluty constraint
	A_eq: array-like (p,n)
		Matrix in left-hand side of equality constraint
	b_eq: array-like (p,)
		Vector in right-hand side of equality constraint
	lb: array-like (n,)
		Vector of lower bounds
	ub: array-like (n,)
		Vector of upper bounds
	Ls: list of array-likes (p,m)
		List of matrices with m columns defining the quadratic constraints
	ys: list of array-likes (m,)
		Centers of the quadratic constraints
	rhos: list of positive floats
		Radii of quadratic constraints
	names: list of strings, optional
		Names for each of the parameters in the space
	kwargs: dict, optional
		Additional parameters to be passed to cvxpy Problem.solve()
	"""
	def __init__(self, A = None, b = None,
		lb = None, ub = None,
		A_eq = None, b_eq = None,
		Ls = None, ys = None, rhos = None,
		names = None, **kwargs):

		self.tol = 1e-6
		# Determine dimension of space
		self._init_dim(lb = lb, ub = ub, A = A, A_eq = A_eq, Ls = Ls)

		# Start setting default values
		self._lb = self._init_lb(lb)
		self._ub = self._init_ub(ub)
		self._A, self._b = self._init_ineq(A, b)
		self._A_eq, self._b_eq = self._init_eq(A_eq, b_eq)
		self._Ls, self._ys, self._rhos = self._init_quad(Ls, ys, rhos)

		self._init_names(names)


		# self.kwargs = merge(DEFAULT_CVXPY_KWARGS, kwargs)

	def __str__(self):
		ret = "<%s on R^%d" % (self.__class__.__name__, len(self))
		if len(self._Ls) > 0:
			ret += "; %d quadratic constraints" % (len(self._Ls),)
		if self._A.shape[0] > 0:
			ret += "; %d linear inequality constraints" % (self._A.shape[0], )
		if self._A_eq.shape[0] > 0:
			ret += "; %d linear equality constraints" % (self._A_eq.shape[0], )
		ret +=">"

		return ret


	################################################################################
	# Initialization helpers
	################################################################################
	def _init_dim(self, lb = None, ub = None, A = None, A_eq = None, Ls = None):
		"""determine the dimension of the space we are working on"""
		if lb is not None:
			if isinstance(lb, (int, float)):
				m = 1
			else:
				m = len(lb)
		elif ub is not None:
			if isinstance(ub, (int, float)):
				m = 1
			else:
				m = len(ub)
		elif A is not None:
			m = len(A[0])
		elif A_eq is not None:
			m = len(A_eq[0])
		elif Ls is not None:
			m = len(Ls[0][0])
		else:
			raise Exception("Could not determine dimension of space")

		self._dimension = m


	def _init_lb(self, lb):
		if lb is None:
			return -np.inf*np.ones(len(self))
		else:
			if isinstance(lb, (int, float)):
				lb = [lb]
			assert len(lb) == len(self), "Lower bound has wrong dimensions"
			return np.array(lb)

	def _init_ub(self, ub):
		if ub is None:
			return np.inf*np.ones(len(self))
		else:
			if isinstance(ub, (int, float)):
				ub = [ub]
			assert len(ub) == len(self), "Upper bound has wrong dimensions"
			return np.array(ub)

	def _init_ineq(self, A, b):
		if A is None and b is None:
			A = np.zeros((0,len(self)))
			b = np.zeros((0,))
		elif A is not None and b is not None:
			A = np.array(A)
			b = np.array(b)
			if len(b.shape) == 0:
				b = b.reshape(1)

			assert len(b.shape) == 1, "b must have only one dimension"

			if len(A.shape) == 1 and len(b) == 1:
				A = A.reshape(1,-1)

			assert A.shape[1] == len(self), "A has wrong number of columns"
			assert A.shape[0] == b.shape[0], "The number of rows of A and b do not match"
		else:
			raise AssertionError("If using inequality constraints, both A and b must be specified")
		return A, b

	def _init_eq(self, A_eq, b_eq):
		if A_eq is None and b_eq is None:
			A_eq = np.zeros((0,len(self)))
			b_eq = np.zeros((0,))
		elif A_eq is not None and b_eq is not None:
			A_eq = np.array(A_eq)
			b_eq = np.array(b_eq)
			if len(b_eq.shape) == 0:
				b_eq = b_eq.reshape(1)

			assert len(b_eq.shape) == 1, "b_eq must have only one dimension"

			if len(A_eq.shape) == 1 and len(b_eq) == 1:
				A_eq = A_eq.reshape(1,-1)

			assert A_eq.shape[1] == len(self), "A_eq has wrong number of columns"
			assert A_eq.shape[0] == b_eq.shape[0], "The number of rows of A_eq and b_eq do not match"
		else:
			raise AssertionError("If using equality constraints, both A_eq and b_eq must be specified")

		return A_eq, b_eq

	def _init_quad(self, Ls, ys, rhos):
		if Ls is None and ys is None and rhos is None:
			_Ls = []
			_ys = []
			_rhos = []
		elif Ls is not None and ys is not None and rhos is not None:
			assert len(Ls) == len(ys) == len(rhos), "Length of all quadratic constraints must be the same"

			_Ls = []
			_ys = []
			_rhos = []
			for L, y, rho in zip(Ls, ys, rhos):
				assert len(L[0]) == len(self), "dimension of L doesn't match the domain"
				assert len(y) == len(self), "Dimension of center doesn't match the domain"
				assert rho > 0, "Radius must be positive"
				_Ls.append(np.array(L))
				_ys.append(np.array(y))
				_rhos.append(rho)
				# TODO: If constraint is rank-1, should we implicitly convert to a linear inequality constriant
		else:
			raise AssertionError("If providing quadratic constraint, each of Ls, ys, and rhos must be defined")
		return _Ls, _ys, _rhos

	################################################################################
	# Simple properties
	################################################################################
	def __len__(self): return self._dimension

	@property
	def lb(self): return self._lb

	@property
	def ub(self): return self._ub

	@property
	def A(self): return self._A

	@property
	def b(self): return self._b

	@property
	def A_eq(self): return self._A_eq

	@property
	def b_eq(self): return self._b_eq

	@property
	def Ls(self): return self._Ls

	@property
	def ys(self): return self._ys

	@property
	def rhos(self): return self._rhos




	################################################################################
	# Normalization
	################################################################################
	def _normalized_domain(self, **kwargs):
		names_norm = [name + ' (normalized)' for name in self.names]

		return LinQuadDomain(lb = self.lb_norm, ub = self.ub_norm, A = self.A_norm, b = self.b_norm,
			A_eq = self.A_eq_norm, b_eq = self.b_eq_norm, Ls = self.Ls_norm, ys = self.ys_norm, rhos = self.rhos_norm,
			names = names_norm, **merge(self.kwargs, kwargs))

	def _isinside(self, X, tol = TOL):
		return self._isinside_bounds(X, tol = tol) & self._isinside_ineq(X, tol = tol) & self._isinside_eq(X, tol = tol) & self._isinside_quad(X, tol = tol)

	def _extent(self, x, p):
		# Check that direction satisfies equality constraints to a tolerance
		if self.A_eq.shape[0] == 0 or np.all(np.abs(self.A_eq.dot(p) ) < self.tol):
			return min(self._extent_bounds(x, p), self._extent_ineq(x, p), self._extent_quad(x, p))
		else:
			return 0.

	################################################################################
	# Convex Solver Functions
	################################################################################

	# def _build_constraints_norm(self, x_norm):
	# 	r""" Build the constraints corresponding to the domain given a vector x
	# 	"""
	# 	constraints = []

	# 	# Numerical issues emerge with unbounded constraints
	# 	I = np.isfinite(self.lb_norm)
	# 	if np.sum(I) > 0:
	# 		constraints.append( self.lb_norm[I] <= x_norm[I])

	# 	I = np.isfinite(self.ub_norm)
	# 	if np.sum(I) > 0:
	# 		constraints.append( x_norm[I] <= self.ub_norm[I])

	# 	if self.A.shape[0] > 0:
	# 		constraints.append( x_norm.__rmatmul__(self.A_norm) <= self.b_norm)
	# 	if self.A_eq.shape[0] > 0:
	# 		constraints.append( x_norm.__rmatmul__(self.A_eq_norm) == self.b_eq_norm)

	# 	for L, y, rho in zip(self.Ls_norm, self.ys_norm, self.rhos_norm):
	# 		if len(L) > 1:
	# 			constraints.append( cp.norm( L @ x_norm - L @ y) <= rho )
	# 		elif len(L) == 1:
	# 			constraints.append( cp.norm(L @ x_norm - L @ y) <= rho)

	# 	return constraints


	# def _build_constraints(self, x):
	# 	r""" Build the constraints corresponding to the domain given a vector x
	# 	"""
	# 	constraints = []

	# 	# Numerical issues emerge with unbounded constraints
	# 	I = np.isfinite(self.lb)
	# 	if np.sum(I) > 0:
	# 		constraints.append( self.lb[I] <= x[I])

	# 	I = np.isfinite(self.ub)
	# 	if np.sum(I) > 0:
	# 		constraints.append( x[I] <= self.ub[I])

	# 	if self.A.shape[0] > 0:
	# 		constraints.append( x.__rmatmul__(self.A) <= self.b)
	# 	if self.A_eq.shape[0] > 0:
	# 		constraints.append( x.__rmatmul__(self.A_eq) == self.b_eq)

	# 	for L, y, rho in zip(self.Ls, self.ys, self.rhos):
	# 		if len(L) > 1:
	# 			constraints.append( cp.norm(L @ x - L.dot(y)) <= rho )
	# 		elif len(L) == 1:
	# 			constraints.append( cp.norm(L @ x - L.dot(y)) <= rho)

	# 	return constraints


	################################################################################
	#
	################################################################################

	def add_constraints(self, A = None, b = None, lb = None, ub = None, A_eq = None, b_eq = None,
		Ls = None, ys = None, rhos = None):
		r"""Add new constraints to the domain
		"""
		lb = self._init_lb(lb)
		ub = self._init_ub(ub)
		A, b = self._init_ineq(A, b)
		A_eq, b_eq = self._init_eq(A_eq, b_eq)
		Ls, ys, rhos = self._init_quad(Ls, ys, rhos)

		# Update constraints
		lb = np.maximum(lb, self.lb)
		ub = np.minimum(ub, self.ub)

		A = np.vstack([self.A, A])
		b = np.hstack([self.b, b])

		A_eq = np.vstack([self.A_eq, A_eq])
		b_eq = np.hstack([self.b_eq, b_eq])

		Ls = self.Ls + Ls
		ys = self.ys + ys
		rhos = self.rhos + rhos

		if len(Ls) > 0:
			return LinQuadDomain(lb = lb, ub = ub, A = A, b = b, A_eq = A_eq, b_eq = b_eq,
				 Ls = Ls, ys = ys, rhos = rhos, names = self.names)
		elif len(b) > 0 or len(b_eq) > 0:
			from .linineq import LinIneqDomain
			return LinIneqDomain(lb = lb, ub = ub, A = A, b = b, A_eq = A_eq, b_eq = b_eq, names = self.names)
		else:
			from .box import BoxDomain
			return BoxDomain(lb = lb, ub = ub, names = self.names)

	def __and__(self, other):
		if isinstance(other, LinQuadDomain) or (isinstance(other, TensorProductDomain) and other.is_linquad_domain):
			return self.add_constraints(lb = other.lb, ub = other.ub,
				A = other.A, b = other.b, A_eq = other.A_eq, b_eq = other.b_eq,
				Ls = other.Ls, ys = other.ys, rhos = other.rhos)
		else:
			raise NotImplementedError

	def __rand__(self, other):
		return self.__and__(other)



import unittest

from context import grama as gr
from context import data

from pandas import Series
from numpy import NaN, arange
from numpy.random import shuffle
from scipy.stats import norm

X = gr.Intention()

##==============================================================================
## mask helper tests
##==============================================================================


class TestFactors(unittest.TestCase):
    def setUp(self):
        pass

    def test_fct_reorder(self):
        ang_fct = gr.fct_reorder(data.df_stang.ang, data.df_stang.E)

        self.assertTrue(list(ang_fct.categories) == [0, 90, 45])

class TestPandasUtils(unittest.TestCase):
    def setUp(self):
        pass

    def test_fillna(self):
        s_nan = Series([NaN] * 2)
        s_filled = gr.fillna(s_nan, value=0)
        s_true = Series([0.0] * 2)

        self.assertTrue((s_filled == s_true).all())

class TestPareto(unittest.TestCase):
    def setUp(self):
        pass

    def test_pareto_min(self):
        df_test = gr.df_make(
            x=[1, 2, 0, 1, 2, 0, 1, 2],
            y=[0, 0, 1, 1, 1, 2, 2, 2],
            p=[1, 0, 1, 0, 0, 0, 0, 0],
        )

        # Test for accuracy
        self.assertTrue(
            (
                df_test
                >> gr.tf_mutate(p_comp=gr.pareto_min(X.x, X.y))
                >> gr.tf_mutate(flag=X.p == X.p_comp)
            ).flag.all()
        )

        # Check for ValueError
        with self.assertRaises(ValueError):
            gr.pareto_min([1], [1, 2, 3])

    def test_stratum_min(self):
        df_test = gr.df_make(
            x=[1, 2, 0, 1, 2, 0, 1, 2],
            y=[0, 0, 1, 1, 1, 2, 2, 2],
            p=[1, 2, 1, 2, 3, 2, 3, 4],
        )

        # Test for accuracy
        self.assertTrue(
            (
                df_test
                >> gr.tf_mutate(p_comp=gr.stratum_min(X.x, X.y))
                >> gr.tf_mutate(flag=X.p == X.p_comp)
            ).flag.all()
        )

        # Check for ValueError
        with self.assertRaises(ValueError):
            gr.stratum_min([1], [1, 2, 3])

class TestQQ(unittest.TestCase):
    def setUp(self):
        pass

    def test_qqvals(self):
        # Set up data
        n = 10
        i = arange(1, n + 1)
        p = (i - 0.3175) / (len(i) + 0.365)
        p[0] = 1 - 0.5**(1/n)
        p[-1] = 0.5**(1/n)
        q = norm.ppf(p)

        # Correct values
        marg = gr.marg_mom("norm", mean=0, sd=1) # Use true distribution
        q_res = gr.qqvals(q, marg=marg)
        self.assertTrue((q == q_res).all())

        # Handles shuffling
        shuffle(q)
        self.assertTrue((q == gr.qqvals(q, marg=marg)).all())

class TestArray(unittest.TestCase):
    def setUp(self):
        pass

    def test_linspace(self):
        # Works in pipeline
        (
            gr.df_make(i=range(10))
            >> gr.tf_mutate(
                x=gr.linspace(0, 1, gr.n(X.index)),
                l=gr.logspace(0, 1, gr.n(X.index)),
            )
        )



from __future__ import division

import numpy as np

# import cvxpy as cp

from .domain import TOL
from .linquad import LinQuadDomain

from ..misc import merge

class LinIneqDomain(LinQuadDomain):
	r"""A domain specified by a combination of linear equality and inequality constraints.

	Here we build a domain specified by three kinds of constraints:
	bound constraints :math:`\text{lb} \le \mathbf{x} \le \text{ub}`,
	inequality constraints :math:`\mathbf{A} \mathbf{x} \le \mathbf{b}`,
	and equality constraints :math:`\mathbf{A}_{\text{eq}} \mathbf{x} = \mathbf{b}_{\text{eq}}`:

	.. math::

		\mathcal{D} := \left \lbrace
			\mathbf{x} : \text{lb} \le \mathbf{x} \le \text{ub}, \
			\mathbf{A} \mathbf{x} \le \mathbf{b}, \
			\mathbf{A}_{\text{eq}} \mathbf{x} = \mathbf{b}_{\text{eq}}
		\right\rbrace \subset \mathbb{R}^m

	Parameters
	----------
	A: array-like (m,n)
		Matrix in left-hand side of inequality constraint
	b: array-like (m,)
		Vector in right-hand side of the ineqaluty constraint
	A_eq: array-like (p,n)
		Matrix in left-hand side of equality constraint
	b_eq: array-like (p,)
		Vector in right-hand side of equality constraint
	lb: array-like (n,)
		Vector of lower bounds
	ub: array-like (n,)
		Vector of upper bounds
	kwargs: dict, optional
		Additional parameters to pass to solvers

	"""
	def __init__(self, A = None, b = None, lb = None, ub = None, A_eq = None, b_eq = None, names = None, **kwargs):
		LinQuadDomain.__init__(self, A = A, b = b, lb = lb, ub = ub, A_eq = A_eq, b_eq = b_eq, names = names, **kwargs)

	def _isinside(self, X, tol = TOL):
		return self._isinside_bounds(X, tol = tol) & self._isinside_ineq(X, tol = tol) & self._isinside_eq(X, tol = tol)

	def _extent(self, x, p):
		return min(self._extent_bounds(x, p), self._extent_ineq(x, p))

	def _normalized_domain(self, **kwargs):
		names_norm = [name + ' (normalized)' for name in self.names]
		return LinIneqDomain(lb = self.lb_norm, ub = self.ub_norm, A = self.A_norm, b = self.b_norm,
			A_eq = self.A_eq_norm, b_eq = self.b_eq_norm, names = names_norm, **merge(self.kwargs, kwargs))


	# def chebyshev_center(self):
	# 	r"""Estimates the Chebyshev center using the constrainted least squares approach

	# 	Solves the linear program finding the radius :math:`r` and Chebyshev center :math:`\mathbf{x}`.

	# 	.. math::

	# 		\max_{r\in \mathbb{R}^+, \mathbf{x} \in \mathcal{D}} &\  r \\
	# 		\text{such that} & \ \mathbf{a}_i^\top \mathbf{x} + r \|\mathbf{a}_i\|_2 \le b_i

	# 	where we have expressed the domain in terms of the linear inequality constraints
	# 	:math:`\mathcal{D}=\lbrace \mathbf{x} : \mathbf{A}\mathbf{x} \le \mathbf{b}\rbrace`
	# 	and :math:`\mathbf{a}_i^\top` are the rows of :math:`\mathbf{A}` as described in [BVNotes]_.


	# 	Returns
	# 	-------
	# 	center: np.ndarray(m,)
	# 		Center of the domain
	# 	radius: float
	# 		radius of largest circle inside the domain

	# 	References
	# 	----------
	# 	.. [BVNotes] https://see.stanford.edu/materials/lsocoee364a/04ConvexOptimizationProblems.pdf, page 4-19.
	# 	"""
	# 	m, n = self.A.shape

	# 	# Merge the bound constraints into A
	# 	A = self.A_aug
	# 	b = self.b_aug

	# 	# See p.4-19 https://see.stanford.edu/materials/lsocoee364a/04ConvexOptimizationProblems.pdf
	# 	#
	# 	normA = np.sqrt( np.sum( np.power(A, 2), axis=1 ) ).reshape((A.shape[0], ))

	# 	r = cp.Variable(1)
	# 	x = cp.Variable(len(self))

	# 	constraints = [A @ x + normA * r <= b]
	# 	if len(self.A_eq) > 0:
	# 		constraints += [self.A_eq @ x == self.b_eq]

	# 	problem = cp.Problem(cp.Maximize(r), constraints)
	# 	problem.solve(**self.kwargs)
	# 	radius = float(r.value)
	# 	center = np.array(x.value).reshape(len(self))

	# 	#AA = np.hstack(( A, normA.reshape(-1,1) ))
	# 	#c = np.zeros((A.shape[1]+1,))
	# 	#c[-1] = -1.0
	# 	#A_eq = np.hstack([self.A_eq, np.zeros( (self.A_eq.shape[0],1))])
	# 	#zc = linprog(c, A_ub = AA, b_ub = b, A_eq = A_eq, b_eq = self.b_eq )
	# 	#center = zc[:-1].reshape((n,))
	# 	#radius = zc[-1]
	# 	#print(center)
	# 	#print(self.isinside(center))
	# 	#zc = cp.Variable(len(self)+1)
	# 	#prob = cp.Problem(cp.Maximize(zc[-1]), [zc.__rmatmul__(AA) <= b])
	# 	#prob.solve()
	# 	#print(zc.value[0:-1])

	# 	self._radius = radius
	# 	self._cheb_center = center

	# 	return center, radius

	@property
	def radius(self):
		try:
			return self._radius
		except:
			self.chebyshev_center()
			return self._radius

	@property
	def center(self):
		try:
			return self._cheb_center
		except:
			self.chebyshev_center()
			return self._cheb_center

	@property
	def Ls(self): return []

	@property
	def ys(self): return []

	@property
	def rhos(self): return []



__all__ = [
    "tran_pivot_longer",
    "tf_pivot_longer",
    "tran_pivot_wider",
    "tf_pivot_wider",
]


import re
from grama import add_pipe, tran_select, symbolic_evaluation, \
    group_delegation, resolve_selection, Intention
from numpy import max as npmax
from numpy import NaN, size, where, zeros
from pandas import DataFrame, IndexSlice, MultiIndex, RangeIndex, Series, \
    concat, isnull, pivot, pivot_table
from pandas.api.types import is_int64_dtype


def tran_pivot_longer (
    df,
    columns,
    index_to = None,
    names_to = None,
    #names_prefix = None,
    names_sep = None,
    names_pattern = None,
    #names_ptypes = list(),
    #names_transform = list(),
    #names_repair,
    values_to = None,
    #values_drop_na = False,
    #values_ptypes = list(),
    #values_transform = list(),
):
    """Lengthen a dataset

    "Lengthens" data by increasing the number of rows and decreasing the
    number of columns.

    Args:
        df (DataFrame): DataFrame passed through
        columns (str): Label of column(s) to pivot into longer format
        index_to(str): str name to create a new representation index of observations; Optional.
        names_to (str): name to use for the 'variable' column, if None frame.columns.name
                        is used or ‘variable’
                          • .value indicates that component of the name defines
                            the name of the column containing the cell values,
                            overriding values_to
        names_sep (str OR list of int): delimter to seperate the values of the argument(s) from
                        the 'columns' parameter into 2 new columns with those
                        values split by that delimeter
                          • Regex expression is a valid input for names_sep
        names_pattern (str): Regular expression with capture groups to define targets for names_to.
        values_to (str): name to use for the 'value' column; overridden if ".value" is provided in names_to argument.

    Notes:
        Only one of names_sep OR names_pattern may be given.

    Returns:
        DataFrame: result of being pivoted into a longer format

    Examples::

        import grama as gr
        ## Simple example
        (
            gr.df_make(
                A=[1, 2, 3],
                B=[4, 5, 6],
                C=[7, 8, 9],
            )
            >> gr.tf_pivot_longer(
                columns=["A", "B", "C"],
                names_to="variable",
                values_to="value",
            )
        )

        ## Matching columns on patterns
        (
            gr.df_make(
                x1=[1, 2, 3],
                x2=[4, 5, 6],
                x3=[7, 8, 9],
            )
            >> gr.tf_pivot_longer(
                columns=gr.matches("\\d+"),
                names_to="variable",
                values_to="value",
            )
        )

        ## Separating column names and data on a names_pattern
        (
            gr.df_make(
                E00=[1, 2, 3],
                E45=[4, 5, 6],
                E90=[7, 8, 9],
            )
            >> gr.tf_pivot_longer(
                columns=gr.matches("\\d+"),
                names_to=[".value", "angle"],
                names_pattern="(E)(\\d+)",
            )
        )

    """


    ########### Pre-Check List #############
    ### Check if tran_select was used
    if isinstance(columns, DataFrame):
        columns = columns.columns.values

    ### Check if selection helper was used:
    if isinstance(columns,Intention):
        columns = pivot_select(df, columns)
        if size(columns) == 0:
            raise ValueError("""Selection helper has found no matches. Revise
                columns input.""")

    ### Check if names_to is a list or str
    names_str = False
    if isinstance(names_to, str):
        names_str = True
        if names_sep is not None:
            raise TypeError("""In order to use names_sep more than 1 value
                needs to passed to names_to""")

    ### Check for .value input
    dot_value = False
    if names_str is False:
        for i, v in enumerate(names_to):
            if names_to[i] == ".value":
                dot_value = True
    else:
        if names_to == ".value":
            dot_value = True

    ### Check values_to argument
    if values_to is None:
        values_to = "values"

    if names_pattern and names_sep:
        raise ValueError("""Both names_sep and names_pattern were used,
            only one or the other is required""")


    #######################################


    ########### .value pivot #############

    ### Check if .value operation needs to occur
    if dot_value is True:

        ### collect unused columns to pivot around
        data_index = collect_indexes(df, columns)

        if names_sep is not None or names_pattern is not None:
            ### Add index and split column to dataset
            longer = df.reset_index().melt(
                    id_vars="index",
                    var_name="split",
                    value_vars=columns,
                    value_name=values_to
                )

            ### DataFrame Cleanup
            longer = split_cleanup(
                longer=longer,
                names_to=names_to,
                names_pattern=names_pattern,
                names_sep=names_sep,
                values_to=values_to
            )

        else:
            ### Add index column and .value column
            longer = df.reset_index().melt(
                    id_vars="index",
                    var_name=".value",
                    value_vars=columns,
                    value_name=values_to
                )

        ### clean up index_to call
        longer = index_to_cleanup(df, longer, data_index)

        ### arrange what indexes_from should be
        if names_str is True:
            indexes = ["index"] + data_index
        else:
            names_to = list(names_to)
            value_loc = names_to.index(".value")
            if value_loc == 0:
                indexes = ["index"] + data_index + names_to[1:]
            else:
                indexes = ["index"] + data_index + names_to[0:value_loc] \
                    + names_to[(value_loc+1):]

        ### Pivot wider the .value column
        value_longer = tran_pivot_wider(
            longer,
            indexes_from=indexes,
            names_from=".value",
            values_from=values_to
        )

        if index_to is None:
            ### drop "index" column
            value_longer.drop("index", axis=1, inplace=True)
        else:
            ### rename index column to desired: index_to
            value_longer.rename(columns={'index': index_to},inplace=True)

        return value_longer

    #########################################


    ########### names_sep pivot #############

    ### Only if names_sep is used
    if names_sep is not None or names_pattern is not None:

        ### collect unused columns to pivot around
        data_index = collect_indexes(df, columns)

        if index_to is None:
            ### initial pivoted DataFrame
            longer = df.reset_index().melt(
                id_vars=data_index,
                var_name="split",
                value_vars=columns,
                value_name=values_to
            )

            ### DataFrame Cleanup
            longer = split_cleanup(
                longer=longer,
                names_to=names_to,
                names_pattern=names_pattern,
                names_sep=names_sep,
                values_to=values_to
            )

            return(longer)

        ### Add index column to dataset
        longer = df.reset_index().melt(
                id_vars="index",
                var_name="split",
                value_vars=columns,
                value_name=values_to
            )
        ### rename index column to desired: index_to
        longer.rename(columns={'index': index_to},inplace=True)

        longer = index_to_cleanup(df, longer, data_index)

        ### DataFrame Cleanup
        longer = split_cleanup(
            longer=longer,
            names_to=names_to,
            names_pattern=names_pattern,
            names_sep=names_sep,
            values_to=values_to
        )

        return(longer)

    ######################################


    ########### normal pivot #############

    ### Check if index_to is provided
    if index_to is None:

        ### check to see if all columns are used already
        data_columns = df.columns.values
        data_index = [x for x in data_columns if x not in columns]

        ### check if data_index is empty and if it has a RangeIndex
        if not data_index:
            if is_int64_dtype(df.index.dtype):
                # if so do not add extra index column and pivot
                longer = df.reset_index().melt(
                    id_vars=None,
                    var_name=names_to,
                    value_vars=columns,
                    value_name=values_to
                )
                return(longer)

            # if it does not have a RangeIndex create new column from ID column
            # and add RangeIndex
            longer = df.reset_index().melt(
                id_vars="index",
                var_name=names_to,
                value_vars=columns,
                value_name=values_to
            )
            return(longer)

        ### look for unused columns to pivot around
        data_used = columns
        data_index = [x for x in data_columns if x not in data_used]

        ### pivot with leftover name that would be the index column
        if data_index:
            longer = df.reset_index().melt(
                id_vars=data_index,
                var_name=names_to,
                value_vars=columns,
                value_name=values_to
            )
            return(longer)

    ### collect unused columns to preserve post pivot
    data_index = collect_indexes(df, columns)

    ### Add index column to dataset
    longer = df.reset_index().melt(
            id_vars="index",
            var_name=names_to,
            value_vars=columns,
            value_name=values_to
        )
    ### rename index column to desired: index_to
    longer.rename(columns={'index': index_to},inplace=True)

    longer = index_to_cleanup(df, longer, data_index)

    return longer

    ######################################

tf_pivot_longer = add_pipe(tran_pivot_longer)


def tran_pivot_wider (
     df,
     #id_cols,
     names_from,
     indexes_from = None,
     #names_prefix,
     #names_sep,
     #names_glue = None,
     #names_sort = False,
     #names_glue,
     values_from = None,
     #values_fill = None,
     #values_fn = None
 ):
    """Widen a dataset

     "Widens" data by increasing the number of columns and decreasing the
     number of rows.

     Args:
         df (DataFrame): DataFrame passed through
         names_from (str): Column(s) name(s) to use to make new columns
         indexes_from (str): Column(s) to use to make new index, if None will
                             preserve and use unspecified columns as index
         values_from (str): Column(s) to use as new values column(s)

     Returns:
        DataFrame: result of being pivoted wider

     Examples::

        import grama as gr
        ## Simple example
        (
            gr.df_make(var=["x", "x", "y", "y"], value=[0, 1, 2, 3])
            >> gr.tf_pivot_wider(
                names_from="var",
                values_from="value",
            )
        (

    """

    if indexes_from is None:
        ### Clean columns list to find unused columns and preserve them
        data_columns = df.columns
        data_values_used = values_from
        data_columns_used = names_from
        data_1st_clean = [x for x in data_columns if x not in data_columns_used]
        data_index_clean = [x for x in data_1st_clean if x not in data_values_used]
        if not data_index_clean:
            data_index_clean = None

        ### try pivot, if no repeats exist in data_index_clean values continue,
        ### else use pivot_table to handle repreated values
        wider = pivot(
            df,
            index=data_index_clean,
            columns=names_from,
            values=values_from
        )
        ### Post cleaning of leftover column and index names
        wider.columns.name = None
        # preserve rows by re-inserting them
        if data_index_clean is not None:
            for i, v in enumerate(data_index_clean):
                wider.insert(
                    loc=i,
                    column=data_index_clean[i],
                    value=wider.index.get_level_values(data_index_clean[i])
                    )
        # remake index to be numbered column with no name
        wider.index = RangeIndex(start=0,stop=len(wider),step=1)
        wider.index.name = None

        return wider

    ### If indexes_from exists do the same repeat check as above but with the
    ### provided indexes
    wider = pivot(
        df,
        index=indexes_from,
        columns=names_from,
        values=values_from
    )
    ### Post cleaning of leftover column and index names
    wider.columns.name = None
    ## preserve rows by re-inserting them
    if isinstance(indexes_from, str):
        wider.insert(0,indexes_from,wider.index)
    # if str, insert as so, else loop through list type of indexes_from
    else:
        for i, v in enumerate(indexes_from):
            wider.insert(
                loc=i,
                column=indexes_from[i],
                value=wider.index.get_level_values(indexes_from[i])
                )
    # remake index to be numbered column with no name
    wider.index = RangeIndex(start=0,stop=len(wider),step=1)
    wider.index.name = None

    return wider

tf_pivot_wider = add_pipe(tran_pivot_wider)


def split_cleanup(
    longer,
    names_to,
    names_pattern,
    names_sep,
    values_to
):
    """
        split_cleanup cleans up pivots that use the names_sep functionality
    """
    ### clean up DataFrame
    # split columns based on character/regex or position via int
    if names_sep:

        # if a positional argument is used
        if isinstance(names_sep, list):
            split = longer.split.str
            names_sep.sort()

            # if one positional split is called
            if len(names_sep) == 1:
                left, right = split[:names_sep[0]], split[names_sep[0]+1:]
                split_columns = DataFrame({
                    0:left,
                    1:right
                })
            # if multiple positional splits are called
            else:
                splits = [] # split container

                # initial split
                left, right = split[:names_sep[0]], split[names_sep[0]+1:]
                splits.append(left)

                # convert and make a pointer of the remaining right half
                point = right.to_frame().split.str

                # iteratively seperate across the remaining names
                for i in range(len(names_sep)-1):
                    sep = i+1                 # current seperator
                    offset = names_sep[sep-1] # previous seperator value

                    left = point[:(names_sep[sep]-offset-1)]  # extra -1 to offset
                    right = point[names_sep[sep]-offset-1+1:] # range(len(x)-1)
                    splits.append(left)
                    point = right.to_frame().split.str # convert and re-point

                splits.append(right) # append final split

                # create final DataFrame of split values
                split_columns = DataFrame()
                for i, name in enumerate(splits):
                    column = name.to_frame()
                    split_columns[i] = column

        # if a string argument is used
        else:
            split_columns = longer.split.str.split(
                names_sep,
                expand=True,
                n=0
            )

    # split columns based on names_pattern regex capture groups
    if names_pattern:
        split = Series(longer.split)
        split_columns = split.str.extract(names_pattern)

    # add back split columns to DataFrame
    longer = concat([longer,split_columns], axis=1)

    # drop column that the split came from
    longer.drop("split", axis=1, inplace=True)

    # rename columns
    for i,v in enumerate(names_to):
        longer.rename(columns={i: v},inplace=True)

    # if any values are None make them NaN
    for i,v in enumerate(names_to):
        for j, w in enumerate(longer[names_to[i]].values):
            if w is None:
                longer[names_to[i]].values[j] = NaN

    # reorder values column to the end
    longer = longer[[c for c in longer if c not in values_to] + [values_to]]

    return longer


def collect_indexes(df, columns):
    """
        collect_indexes finds unused column to pivot around
    """
    ### look for unused columns to pivot around
    data_used = columns
    data_columns = df.columns.values
    data_index = [x for x in data_columns if x not in data_used]

    return data_index


def index_to_cleanup(df, longer, data_index):
    """
        index_to_cleanup cleans up longer if index_to was called to the function
    """
    ### if there was columns needing to be re-inserted do so
    if data_index is not None:
        for i, v in enumerate(data_index):
            longer.insert(
                loc=i+1,
                column=data_index[i],
                value=df[data_index[i]]
            )

    ### repair NaN values from transformation
    for index in data_index:
        length = len(df[index])
        for i, v in enumerate(longer[index]):
            if isnull(longer[index].values[i]):
                longer[index].values[i] = longer[index].values[i%length]

    return longer


@group_delegation
@symbolic_evaluation(eval_as_selector=True)
def pivot_select(df, columns):
    """
        pivot_select helps resolve the use of a selection helper in the columns
        argument (e.g: columns = gr.matches("\\d+"))
    """
    ordering, column_indices = resolve_selection(df, columns)
    if (column_indices == 0).all():
        return df[[]]
    selection = where(
        (column_indices == npmax(column_indices)) & (column_indices >= 0)
    )[0]
    df = df.iloc[:, selection]

    return df.columns.values



import unittest
import io
import sys

from context import grama as gr
from numpy import eye, zeros, sqrt
from numpy.random import multivariate_normal
from pandas import DataFrame

DF = gr.Intention()

## Test support points
##################################################
class TestReweight(unittest.TestCase):
    def setUp(self):
        self.md = (
            gr.Model()
            >> gr.cp_function(
                fun=lambda x: x,
                var=["x"],
                out=["y"],
            )
            >> gr.cp_marginals(x=dict(dist="norm", loc=0, scale=1))
            >> gr.cp_copula_independence()
        )

    def test_tran_reweight(self):
        """Test the functionality of tran_reweight()

        """
        ## Correctness
        # Choose scale based on Owen (2013) Exercise 9.7
        md_new = (
            self.md
            >> gr.cp_marginals(x=dict(dist="norm", loc=0, scale=sqrt(4/5)))
        )

        df_base = (
            self.md
            >> gr.ev_sample(n=500, df_det="nom", seed=101)
        )

        df = (
            df_base
            >> gr.tf_reweight(md_base=self.md, md_new=md_new)
            >> gr.tf_summarize(
                mu=gr.mean(DF.y * DF.weight),
                se=gr.sd(DF.y * DF.weight) / gr.sqrt(gr.n(DF.weight)),
                se_orig=gr.sd(DF.y) / gr.sqrt(gr.n(DF.weight)),
            )
        )
        mu = df.mu[0]
        se = df.se[0]
        se_orig = df.se_orig[0]

        self.assertTrue(
            mu - se * 2 < 0 and
            0 < mu + se * 2
        )

        ## Optimized IS should be more precise than ordinary monte carlo
        # print("se_orig = {0:4.3f}".format(se_orig))
        # print("se      = {0:4.3f}".format(se))
        self.assertTrue(se < se_orig)

        ## Invariants
        # Missing input in data
        with self.assertRaises(ValueError):
            gr.tran_reweight(df_base[["y"]], md_base=self.md, md_new=self.md)
        # Input mismatch
        with self.assertRaises(ValueError):
            gr.tran_reweight(
                df_base,
                md_base=self.md,
                md_new=gr.Model()
            )
        # Weights collision
        with self.assertRaises(ValueError):
            gr.tran_reweight(
                df_base
                >> gr.tf_mutate(weight=0),
                md_base=self.md,
                md_new=self.md
            )
